import logging
import logging.config
from typing import Final

from investir.config import config


class CustomFormatter(logging.Formatter):
    CSI: Final = "\033["
    BRIGHT_CYAN: Final = f"{CSI}96m"
    BOLD_WHITE: Final = f"{CSI}1;37m"
    BRIGHT_YELLOW: Final = f"{CSI}93m"
    RED: Final = f"{CSI}31m"
    BOLD_RED: Final = f"{CSI}1;31m"
    RESET: Final = f"{CSI}0m"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        fmt = kwargs.get("fmt", "%(message)s")

        self.formats = {
            logging.DEBUG: self.BRIGHT_CYAN + fmt + self.RESET,
            logging.INFO: self.BOLD_WHITE + fmt + self.RESET,
            logging.WARNING: self.BRIGHT_YELLOW + fmt + self.RESET,
            logging.ERROR: self.RED + fmt + self.RESET,
            logging.CRITICAL: self.BOLD_RED + fmt + self.RESET,
        }

    def format(self, record: logging.LogRecord) -> str:
        log_fmt = self.formats.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


def configure_logger() -> None:
    logging.config.dictConfig(
        {
            "version": 1,
            "disable_existing_loggers": True,
            "formatters": {
                "standard": {
                    "()": CustomFormatter if config.use_colour else logging.Formatter,
                    "format": "%(levelname)8s | %(message)s",
                }
            },
            "handlers": {
                "console": {
                    "class": "logging.StreamHandler",
                    "formatter": "standard",
                },
            },
            "loggers": {
                "root": {"level": "NOTSET", "handlers": ["console"]},
                "investir": {"level": config.log_level},
                "yfinance": {"level": logging.CRITICAL},
            },
        }
    )



import datetime
import logging
from collections.abc import Callable, Iterable, Mapping, Sequence
from decimal import Decimal
from typing import Final

from investir.config import config
from investir.typing import Year

logger = logging.getLogger(__name__)

TAX_YEAR_MONTH: Final = 4
TAX_YEAR_START_DAY: Final = 6
TAX_YEAR_END_DAY: Final = 5


def tax_year_period(tax_year: Year) -> tuple[datetime.date, datetime.date]:
    tax_year_start = datetime.date(tax_year, TAX_YEAR_MONTH, TAX_YEAR_START_DAY)
    tax_year_end = datetime.date(tax_year + 1, TAX_YEAR_MONTH, TAX_YEAR_END_DAY)
    return tax_year_start, tax_year_end


def date_to_tax_year(date: datetime.date) -> Year:
    tax_year_start, _ = tax_year_period(Year(date.year))
    if date >= tax_year_start:
        return Year(tax_year_start.year)
    return Year(tax_year_start.year - 1)


def tax_year_short_date(tax_year: Year) -> str:
    return f"{tax_year}/{(tax_year + 1) % 100}"


def tax_year_full_date(tax_year: Year) -> str:
    return f"6th April {tax_year} to 5th April {tax_year + 1}"


def multifilter(filters: Sequence[Callable] | None, iterable: Iterable) -> Iterable:
    if not filters:
        return iterable
    return filter(lambda x: all(f(x) for f in filters), iterable)


def raise_or_warn(ex: Exception) -> None:
    if config.strict:
        raise ex
    logger.warning(ex)


def read_decimal(val: str, default: Decimal = Decimal("0.0")) -> Decimal:
    return Decimal(val) if val.strip() else default


def dict2str(d: Mapping[str, str]) -> str:
    return str({k: v for k, v in d.items() if v.strip()})


def boldify(text: str) -> str:
    return f"\033[1m{text}\033[0m"


def unboldify(text: str) -> str:
    text = text.replace("\033[1m", "")
    text = text.replace("\033[0m", "")
    return text



import importlib.metadata
import logging
import operator
from collections.abc import Callable, Sequence
from datetime import datetime
from pathlib import Path
from typing import Annotated, Optional

import click
import typer

from investir.config import config
from investir.exceptions import InvestirError
from investir.findata import (
    FinancialData,
    YahooFinanceExchangeRateProvider,
    YahooFinanceSecurityInfoProvider,
)
from investir.logging import configure_logger
from investir.parser import ParserFactory
from investir.prettytable import OutputFormat
from investir.taxcalculator import TaxCalculator
from investir.transaction import Acquisition, Disposal, Transaction
from investir.trhistory import TrHistory
from investir.typing import Ticker, Year
from investir.utils import boldify, tax_year_full_date, tax_year_short_date

logger = logging.getLogger(__name__)


class OrderedCommands(typer.core.TyperGroup):
    def list_commands(self, ctx: click.Context) -> list[str]:
        return list(self.commands.keys())


class MutuallyExclusiveOption(click.exceptions.UsageError):
    def __init__(self, opt1: str, opt2: str) -> None:
        super().__init__(f"Option {opt1} cannot be used together with option {opt2}")


app = typer.Typer(
    cls=OrderedCommands,
    context_settings={"help_option_names": ["-h", "--help"]},
    no_args_is_help=True,
    pretty_exceptions_enable=False,
)

FilesArg = Annotated[
    list[Path],
    typer.Argument(
        exists=True,
        dir_okay=False,
        readable=True,
        help="CSV files with the account activity.",
        show_default=False,
    ),
]

TaxYearOpt = Annotated[
    Optional[int],
    typer.Option(
        min=2008,
        max=datetime.now().year,
        metavar="TAX-YEAR",
        help="Filter by tax year.",
        show_default=False,
    ),
]

TickerOpt = Annotated[
    Optional[str],
    typer.Option(metavar="TICKER", help="Filter by ticker.", show_default=False),
]


OutputFormatOpt = Annotated[
    OutputFormat,
    typer.Option("--output", "-o", help="Output format."),
]


def abort(message: str) -> None:
    logger.critical(message)
    raise typer.Exit(code=1)


def parse(input_files: list[Path]) -> tuple[TrHistory, TaxCalculator]:
    orders = []
    dividends = []
    transfers = []
    interest = []

    for path in input_files:
        logger.info("Parsing input file: %s", path)
        if parser := ParserFactory.create_parser(path):
            try:
                result = parser.parse()
            except InvestirError as ex:
                abort(str(ex))
            logger.info(
                "Parsed: "
                "%s orders, %s dividend payments, %s transfers, %s interest payments",
                len(result.orders),
                len(result.dividends),
                len(result.transfers),
                len(result.interest),
            )

            orders += result.orders
            dividends += result.dividends
            transfers += result.transfers
            interest += result.interest
        else:
            abort(f"Unable to find a parser for {path}")

    tr_hist = TrHistory(
        orders=orders, dividends=dividends, transfers=transfers, interest=interest
    )

    logger.info(
        "Total: %s orders, %s dividend payments, %s transfers, %s interest payments",
        len(tr_hist.orders),
        len(tr_hist.dividends),
        len(tr_hist.transfers),
        len(tr_hist.interest),
    )

    security_info_provider = None
    exchange_rate_provider = None
    if not config.offline:
        security_info_provider = YahooFinanceSecurityInfoProvider()
        exchange_rate_provider = YahooFinanceExchangeRateProvider()

    financial_data = FinancialData(
        security_info_provider,
        exchange_rate_provider,
        tr_hist,
        config.cache_file,
    )
    tax_calculator = TaxCalculator(tr_hist, financial_data)

    return tr_hist, tax_calculator


def create_filters(
    tax_year: int | None = None,
    ticker: str | None = None,
    tr_type: type[Transaction] | None = None,
    amount_op: Callable | None = None,
) -> Sequence[Callable[[Transaction], bool]]:
    filters = []

    if tax_year is not None:
        filters.append(lambda tr: tr.tax_year() == Year(tax_year))

    if ticker is not None:
        filters.append(lambda tr: tr.ticker == ticker)

    if tr_type is not None:
        filters.append(lambda tr: isinstance(tr, tr_type))

    if amount_op is not None:
        filters.append(lambda tr: amount_op(tr.amount, 0.0))

    return filters


def version_callback(value: bool) -> None:
    if value:
        print(f"{__package__} {importlib.metadata.version(__package__)}")
        raise typer.Exit()


@app.callback()
def main_callback(  # noqa: PLR0913
    strict: Annotated[
        bool, typer.Option(help="Abort if data integrity issues are found.")
    ] = config.strict,
    offline: Annotated[
        bool,
        typer.Option(
            "--offline",
            help=(
                "Disable fetching additional data about securities (e.g. list of "
                "share sub-division events) from the Internet."
            ),
        ),
    ] = config.offline,
    cache_file: Annotated[
        Path,
        typer.Option(
            dir_okay=False,
            help=(
                "Cache file to store additional data about securities fetched "
                "from the Internet."
            ),
        ),
    ] = config.cache_file,
    include_fx_fees: Annotated[
        bool, typer.Option(help="Include foreign exchange fees as an allowable cost.")
    ] = config.include_fx_fees,
    verbose: Annotated[
        bool, typer.Option("--verbose", help="Enable additional logging.")
    ] = False,
    quiet: Annotated[
        bool, typer.Option("--quiet", help="Disable all non-critical logging.")
    ] = False,
    colour: Annotated[
        bool, typer.Option(help="Show coloured output.")
    ] = config.use_colour,
    version: Annotated[
        Optional[bool],
        typer.Option(
            "--version",
            callback=version_callback,
            help="Show version information and exit.",
        ),
    ] = None,
) -> None:
    if verbose and quiet:
        raise MutuallyExclusiveOption("--verbose", "--quiet")

    config.strict = strict
    config.offline = offline
    config.cache_file = cache_file
    config.include_fx_fees = include_fx_fees

    if quiet:
        config.log_level = logging.CRITICAL

    if verbose:
        config.log_level = logging.DEBUG

    config.use_colour = colour

    configure_logger()


@app.command("orders")
def orders_command(  # noqa: PLR0913
    files: FilesArg,
    tax_year: TaxYearOpt = None,
    ticker: TickerOpt = None,
    acquisitions_only: Annotated[
        bool, typer.Option("--acquisitions", help="Show only acquisitions.")
    ] = False,
    disposals_only: Annotated[
        bool, typer.Option("--disposals", help="Show only disposals.")
    ] = False,
    format: OutputFormatOpt = OutputFormat.TEXT,
) -> None:
    """
    Show share buy/sell orders.
    """
    if acquisitions_only and disposals_only:
        raise MutuallyExclusiveOption("--acquisitions", "--disposals")

    tr_hist, _ = parse(files)

    tr_type: type[Transaction] | None = None
    if acquisitions_only:
        tr_type = Acquisition
    elif disposals_only:
        tr_type = Disposal

    filters = create_filters(tax_year=tax_year, ticker=ticker, tr_type=tr_type)
    if table := tr_hist.get_orders_table(filters):
        print(table.to_string(format, leading_nl=config.logging_enabled))


@app.command("dividends")
def dividends_command(
    files: FilesArg,
    tax_year: TaxYearOpt = None,
    ticker: TickerOpt = None,
    format: OutputFormatOpt = OutputFormat.TEXT,
) -> None:
    """
    Show share dividends paid out.
    """
    tr_hist, _ = parse(files)
    filters = create_filters(tax_year=tax_year, ticker=ticker)
    if table := tr_hist.get_dividends_table(filters):
        print(table.to_string(format, leading_nl=config.logging_enabled))


@app.command("transfers")
def transfers_command(
    files: FilesArg,
    tax_year: TaxYearOpt = None,
    deposits_only: Annotated[
        bool, typer.Option("--deposits", help="Show only deposits.")
    ] = False,
    withdrawals_only: Annotated[
        bool, typer.Option("--withdrawals", help="Show only withdrawals.")
    ] = False,
    format: OutputFormatOpt = OutputFormat.TEXT,
) -> None:
    """
    Show cash deposits and cash withdrawals.
    """
    if deposits_only and withdrawals_only:
        raise MutuallyExclusiveOption("--deposits", "--withdrawals")

    tr_hist, _ = parse(files)

    if deposits_only:
        amount_op = operator.gt
    elif withdrawals_only:
        amount_op = operator.lt
    else:
        amount_op = None

    filters = create_filters(tax_year=tax_year, amount_op=amount_op)
    if table := tr_hist.get_transfers_table(filters):
        print(table.to_string(format, leading_nl=config.logging_enabled))


@app.command("interest")
def interest_command(
    files: FilesArg,
    tax_year: TaxYearOpt = None,
    format: OutputFormatOpt = OutputFormat.TEXT,
) -> None:
    """
    Show interest earned on cash.
    """
    tr_hist, _ = parse(files)
    filters = create_filters(tax_year=tax_year)
    if table := tr_hist.get_interest_table(filters):
        print(table.to_string(format, leading_nl=config.logging_enabled))


@app.command("capital-gains")
def capital_gains_command(  # noqa: PLR0913
    files: FilesArg,
    gains_only: Annotated[
        bool, typer.Option("--gains", help="Show only capital gains.")
    ] = False,
    losses_only: Annotated[
        bool, typer.Option("--losses", help="Show only capital losses.")
    ] = False,
    tax_year: TaxYearOpt = None,
    ticker: TickerOpt = None,
    format: OutputFormatOpt = OutputFormat.TEXT,
) -> None:
    """
    Show capital gains report.
    """
    if gains_only and losses_only:
        raise MutuallyExclusiveOption("--gains", "--losses")

    if format != OutputFormat.TEXT and tax_year is None:
        raise click.exceptions.UsageError(
            f"The {format.value} format requires the option --tax-year to be used"
        )

    _, tax_calculator = parse(files)
    tax_year = Year(tax_year) if tax_year else None
    ticker = Ticker(ticker) if ticker else None

    try:
        tax_years = sorted(tax_calculator.disposal_years())
    except InvestirError as ex:
        abort((str(ex)))

    if tax_year is not None:
        tax_years = [tax_year]

    for tax_year_idx, tax_year in enumerate(tax_years):
        table, summary = tax_calculator.get_capital_gains_table(
            tax_year, ticker, gains_only, losses_only
        )

        if table:
            print(end="\n" if tax_year_idx == 0 and config.logging_enabled else "")

            if format == OutputFormat.TEXT:
                print(
                    boldify(f"Capital Gains Tax Report {tax_year_short_date(tax_year)}")
                )
                print(tax_year_full_date(tax_year))
                print(table.to_string(format))
                print(summary)
            else:
                print(table.to_string(format, leading_nl=False))


@app.command("holdings")
def holdings_command(
    files: FilesArg,
    ticker: TickerOpt = None,
    show_gain_loss: Annotated[
        bool, typer.Option("--show-gain-loss", help="Show unrealised gain/loss.")
    ] = False,
    format: OutputFormatOpt = OutputFormat.TEXT,
) -> None:
    """
    Show current holdings.
    """
    _, tax_calculator = parse(files)
    ticker = Ticker(ticker) if ticker else None
    if table := tax_calculator.get_holdings_table(ticker, show_gain_loss):
        print(table.to_string(format, leading_nl=config.logging_enabled))


def main() -> None:
    app()



from collections.abc import Mapping
from decimal import Decimal
from pathlib import Path

from investir.typing import ISIN, Ticker


class InvestirError(Exception):
    pass


class ParseError(InvestirError):
    def __init__(self, file: Path, row: Mapping[str, str], message: str) -> None:
        super().__init__(f"{file}: {message} on row {row}")


class TransactionTypeError(ParseError):
    def __init__(self, file: Path, row: Mapping[str, str], tr_type: str) -> None:
        super().__init__(file, row, f"Invalid type of transaction '({tr_type})'")


class CurrencyError(ParseError):
    def __init__(self, file: Path, row: Mapping[str, str], currency: str) -> None:
        super().__init__(file, row, f"Currency not supported ('{currency}')")


class CalculatedAmountError(ParseError):
    def __init__(
        self,
        file: Path,
        row: Mapping[str, str],
        csv_amount: Decimal,
        cal_amount: Decimal,
    ) -> None:
        super().__init__(
            file,
            row,
            f"Calculated amount (£{cal_amount}) is different than the "
            f"expected value (£{csv_amount})",
        )


class FeesError(ParseError):
    def __init__(self, file: Path, row: Mapping[str, str]) -> None:
        super().__init__(file, row, "Stamp duty and conversion fees are both non-zero")


class OrderDateError(ParseError):
    def __init__(self, file: Path, row: Mapping[str, str]) -> None:
        super().__init__(
            file, row, "Orders executed before 6 April of 2008 are not supported"
        )


class IncompleteRecordsError(InvestirError):
    def __init__(self, isin: ISIN, name: str) -> None:
        super().__init__(
            f"Records appear to be incomplete for {name} ({isin}): "
            f"share quantity cannot be negative"
        )


class AmbiguousTickerError(InvestirError):
    def __init__(self, ticker: Ticker) -> None:
        super().__init__(f"Ticker {ticker} is ambiguous (used on different securities)")



import csv
import logging
from collections.abc import Mapping
from datetime import timezone
from decimal import Decimal
from pathlib import Path
from typing import Final

from dateutil.parser import parse as parse_timestamp

from investir.config import config
from investir.const import MIN_TIMESTAMP
from investir.exceptions import (
    CalculatedAmountError,
    CurrencyError,
    FeesError,
    OrderDateError,
    ParseError,
    TransactionTypeError,
)
from investir.parser.factory import ParserFactory
from investir.parser.types import ParsingResult
from investir.transaction import (
    Acquisition,
    Disposal,
    Dividend,
    Interest,
    Order,
    Transfer,
)
from investir.typing import ISIN, Ticker
from investir.utils import dict2str, raise_or_warn, read_decimal

logger = logging.getLogger(__name__)


@ParserFactory.register("Trading212")
class Trading212Parser:
    INITIAL_FIELDS: Final = (
        "Action",
        "Time",
        "ISIN",
        "Ticker",
        "Name",
        "No. of shares",
        "Price / share",
        "Currency (Price / share)",
        "Exchange rate",
    )

    MANDATORY_FIELDS: Final = (
        "Currency (Result)",
        "Total",
        "Currency (Total)",
        "Notes",
        "ID",
        "Currency conversion fee",
        "Currency (Currency conversion fee)",
    )

    OPTIONAL_FIELDS: Final = (
        "Result",
        "Withholding tax",
        "Currency (Withholding tax)",
        "Currency conversion from amount",
        "Currency (Currency conversion from amount)",
        "Currency conversion to amount",
        "Currency (Currency conversion to amount)",
        "Finra fee (GBP)",
        "Stamp duty (GBP)",
    )

    def __init__(self, csv_file: Path) -> None:
        self._csv_file = csv_file
        self._orders: list[Order] = []
        self._dividends: list[Dividend] = []
        self._transfers: list[Transfer] = []
        self._interest: list[Interest] = []

    def can_parse(self) -> bool:
        with self._csv_file.open(encoding="utf-8") as file:
            reader = csv.DictReader(file)
            fieldnames = reader.fieldnames or []
            idx = len(self.INITIAL_FIELDS)
            fields1 = fieldnames[:idx]
            fields2 = fieldnames[idx:]

            if tuple(fields1) != self.INITIAL_FIELDS:
                return False

            if any(f not in fields2 for f in self.MANDATORY_FIELDS):
                return False

            if any(
                f not in self.MANDATORY_FIELDS and f not in self.OPTIONAL_FIELDS
                for f in fields2
            ):
                return False

        return True

    def parse(self) -> ParsingResult:
        parse_fn = {
            "Market buy": self._parse_order,
            "Limit buy": self._parse_order,
            "Market sell": self._parse_order,
            "Limit sell": self._parse_order,
            "Dividend (Ordinary)": self._parse_dividend,
            "Dividend (Dividend)": self._parse_dividend,
            "Dividend (Dividends paid by us corporations)": self._parse_dividend,
            "Dividend (Dividends paid by foreign corporations)": self._parse_dividend,
            "Deposit": self._parse_transfer,
            "Withdrawal": self._parse_transfer,
            "Interest on cash": self._parse_interest,
            "Currency conversion": None,
        }

        with self._csv_file.open(encoding="utf-8") as file:
            for row in csv.DictReader(file):
                tr_type = row["Action"]

                if tr_type not in parse_fn:
                    raise_or_warn(TransactionTypeError(self._csv_file, row, tr_type))
                    continue

                if fn := parse_fn.get(tr_type):
                    currency_total = row["Currency (Total)"]
                    if currency_total != "GBP":
                        raise_or_warn(
                            CurrencyError(self._csv_file, row, currency_total)
                        )
                    fn(row)

        return ParsingResult(
            self._orders, self._dividends, self._transfers, self._interest
        )

    def _parse_order(self, row: Mapping[str, str]) -> None:
        action = row["Action"]
        timestamp = parse_timestamp(row["Time"])
        isin = row["ISIN"]
        ticker = row["Ticker"]
        name = row["Name"]
        num_shares = Decimal(row["No. of shares"])
        price_share = Decimal(row["Price / share"])
        exchange_rate = Decimal(row["Exchange rate"])
        total = Decimal(row["Total"])
        tr_id = row["ID"]
        fx_conversion_fee = read_decimal(row["Currency conversion fee"])
        currency_fx_conversion_fee = row["Currency (Currency conversion fee)"]
        stamp_duty = read_decimal(row.get("Stamp duty (GBP)", ""))
        finra_fee = read_decimal(row.get("Finra fee (GBP)", ""))

        if timestamp.tzinfo is None:
            timestamp = timestamp.replace(tzinfo=timezone.utc)

        if timestamp < MIN_TIMESTAMP:
            raise OrderDateError(self._csv_file, row)

        if stamp_duty and (fx_conversion_fee or finra_fee):
            raise FeesError(self._csv_file, row)

        if fx_conversion_fee and currency_fx_conversion_fee != "GBP":
            raise CurrencyError(self._csv_file, row, currency_fx_conversion_fee)

        order_class: type[Order] = Acquisition
        fees = stamp_duty + finra_fee + fx_conversion_fee

        if action in ("Market sell", "Limit sell"):
            order_class = Disposal
            fees *= -1

        calculated_amount = round(
            (round(price_share * num_shares, 2) / exchange_rate) + fees, 2
        )

        if abs(calculated_amount - total) > Decimal("0.01"):
            raise_or_warn(
                CalculatedAmountError(self._csv_file, row, total, calculated_amount)
            )

        if config.include_fx_fees:
            allowable_fees = abs(fees)
        else:
            allowable_fees = stamp_duty + finra_fee

        self._orders.append(
            order_class(
                timestamp,
                isin=ISIN(isin),
                ticker=Ticker(ticker),
                name=name,
                amount=total - fees,
                quantity=num_shares,
                fees=allowable_fees,
                tr_id=tr_id,
            )
        )

        logger.debug("Parsed row %s as %s\n", dict2str(row), self._orders[-1])

    def _parse_dividend(self, row: Mapping[str, str]):
        timestamp = parse_timestamp(row["Time"])
        isin = row["ISIN"]
        ticker = row["Ticker"]
        name = row["Name"]
        currency_price_share = row["Currency (Price / share)"]
        total = Decimal(row["Total"])
        currency_withholding_tax = row["Currency (Withholding tax)"]
        tr_id = row["ID"]
        fx_conversion_fee = row["Currency conversion fee"]

        if timestamp.tzinfo is None:
            timestamp = timestamp.replace(tzinfo=timezone.utc)

        if fx_conversion_fee:
            raise ParseError(self._csv_file, row, "Dividend with conversion fee")

        if currency_price_share != currency_withholding_tax:
            raise ParseError(
                self._csv_file,
                row,
                "Currency is different for share price and tax withheld",
            )

        self._dividends.append(
            Dividend(
                timestamp,
                isin=ISIN(isin),
                ticker=Ticker(ticker),
                name=name,
                amount=total,
                withheld=None,
                tr_id=tr_id,
            )
        )

        logger.debug("Parsed row %s as %s\n", dict2str(row), self._dividends[-1])

    def _parse_transfer(self, row: Mapping[str, str]):
        action = row["Action"]
        timestamp = parse_timestamp(row["Time"])
        total = Decimal(row["Total"])
        tr_id = row["ID"]

        if timestamp.tzinfo is None:
            timestamp = timestamp.replace(tzinfo=timezone.utc)

        if action == "Withdrawal":
            total *= -1

        self._transfers.append(Transfer(timestamp, tr_id=tr_id, amount=total))

        logger.debug("Parsed row %s as %s\n", dict2str(row), self._transfers[-1])

    def _parse_interest(self, row: Mapping[str, str]):
        timestamp = parse_timestamp(row["Time"])
        total = Decimal(row["Total"])
        tr_id = row["ID"]

        if timestamp.tzinfo is None:
            timestamp = timestamp.replace(tzinfo=timezone.utc)

        self._interest.append(Interest(timestamp, tr_id=tr_id, amount=total))

        logger.debug("Parsed row %s as %s\n", dict2str(row), self._interest[-1])



import datetime
from decimal import Decimal

import pytest

from investir.config import config
from investir.exceptions import InvestirError
from investir.typing import Year
from investir.utils import (
    boldify,
    date_to_tax_year,
    dict2str,
    multifilter,
    raise_or_warn,
    read_decimal,
    tax_year_full_date,
    tax_year_period,
    tax_year_short_date,
    unboldify,
)


def test_tax_year_period():
    tax_year_start, tax_year_end = tax_year_period(2023)
    assert tax_year_start == datetime.date(2023, 4, 6)
    assert tax_year_end == datetime.date(2024, 4, 5)


def test_date_to_tax_year():
    assert date_to_tax_year(datetime.date(2023, 4, 5)) == 2022
    assert date_to_tax_year(datetime.date(2023, 4, 6)) == 2023
    assert date_to_tax_year(datetime.date(2024, 4, 5)) == 2023
    assert date_to_tax_year(datetime.date(2024, 4, 6)) == 2024


def test_tax_year_short_date():
    assert tax_year_short_date(Year(2023)) == "2023/24"


def test_tax_year_full_date():
    assert tax_year_full_date(Year(2023)) == "6th April 2023 to 5th April 2024"


def test_multifilter():
    nums = [1, 2, 3, 4, 5, 6]

    nums_filtered = multifilter(None, nums)
    assert nums_filtered == nums

    nums_filtered = multifilter([], nums)
    assert nums_filtered == nums

    nums_filtered = multifilter([lambda x: x % 2 == 0, lambda x: x > 3], nums)
    assert list(nums_filtered) == [4, 6]


def test_raise_or_warn():
    config.strict = True
    with pytest.raises(InvestirError):
        raise_or_warn(InvestirError("Error"))

    config.strict = False
    raise_or_warn(InvestirError("Error"))


def test_read_decimal():
    assert read_decimal("") == Decimal("0.0")
    assert read_decimal(" ") == Decimal("0.0")
    assert read_decimal("0.0") == Decimal("0.0")
    assert read_decimal("2.5") == Decimal("2.5")
    assert read_decimal("", Decimal("2.7")) == Decimal("2.7")


def test_dict2str():
    d = {"a": "A", "b": "", "c": "C", "d": " "}
    assert dict2str(d) == "{'a': 'A', 'c': 'C'}"


def test_boldify_and_unboldify():
    bold_foo = boldify("foo")
    assert bold_foo == "\x1b[1mfoo\x1b[0m"
    assert unboldify(bold_foo) == "foo"



import logging
from collections import defaultdict, namedtuple
from collections.abc import Callable, Mapping, Sequence
from dataclasses import dataclass
from datetime import date, timedelta
from decimal import Decimal
from typing import TypeAlias

from investir.exceptions import AmbiguousTickerError, IncompleteRecordsError
from investir.findata import FinancialData
from investir.prettytable import PrettyTable
from investir.transaction import Acquisition, Disposal, Order
from investir.trhistory import TrHistory
from investir.typing import ISIN, Ticker, Year
from investir.utils import raise_or_warn

logger = logging.getLogger(__name__)


GroupKey = namedtuple("GroupKey", ["isin", "date", "type"])
GroupDict: TypeAlias = Mapping[GroupKey, Sequence[Order]]


def same_day_match(ord1: Acquisition, ord2: Disposal) -> bool:
    assert ord1.isin == ord2.isin
    return ord1.date == ord2.date


def thirty_days_match(ord1: Acquisition, ord2: Disposal) -> bool:
    assert ord1.isin == ord2.isin
    return ord2.date < ord1.date <= ord2.date + timedelta(days=30)


def gbp(amount):
    sign = "" if amount >= 0.0 else "-"
    return f"{sign}£{abs(amount):.2f}"


@dataclass
class CapitalGain:
    disposal: Disposal
    cost: Decimal
    acquisition_date: date | None = None

    @property
    def gain_loss(self) -> Decimal:
        # Disposal fees have been added to the `cost` so the gross
        # proceeds are used.
        return self.disposal.amount - self.cost

    @property
    def quantity(self) -> Decimal:
        if self.disposal.original_quantity is not None:
            return self.disposal.original_quantity
        return self.disposal.quantity

    @property
    def identification(self) -> str:
        match self.acquisition_date:
            case None:
                return "Section 104"
            case self.disposal.date:
                return "Same day"
            case _:
                return f"Bed & B. ({self.acquisition_date})"

    def __str__(self) -> str:
        return (
            f"{self.disposal.date} "
            f"{self.disposal.isin:<4} "
            f"quantity: {self.quantity}, "
            f"cost: £{self.cost:.2f}, proceeds: £{self.disposal.amount}, "
            f"gain: £{self.gain_loss:.2f}, "
            f"identification: {self.identification}"
        )


@dataclass
class CapitalGainsSummary:
    num_disposals: int
    disposal_proceeds: Decimal
    total_cost: Decimal
    total_gains: Decimal
    total_losses: Decimal

    @property
    def net_gains(self) -> Decimal:
        return self.total_gains - self.total_losses

    def __str__(self) -> str:
        return (
            f"{'Number of disposals:':40}{self.num_disposals:>10}      "
            f"{'Gains in the year, before losses:':34}{gbp(self.total_gains):>10}\n"
            f"{'Disposal proceeds:':40}{gbp(self.disposal_proceeds):>10}      "
            f"{'Losses in the year:':34}{gbp(self.total_losses):>10}\n"
            f"{'Allowable costs (incl. purchase price):':40}"
            f"{gbp(self.total_cost):>10}      "
            f"{'Net gain or loss:':34}{gbp(self.net_gains):>10}\n"
        )


@dataclass
class Section104Holding:
    quantity: Decimal
    cost: Decimal

    def __init__(self, _date: date, quantity: Decimal, cost: Decimal):
        self.quantity = quantity
        self.cost = cost

    def increase(self, _date: date, quantity: Decimal, cost: Decimal) -> None:
        self.quantity += quantity
        self.cost += cost

    def decrease(self, _date: date, quantity: Decimal, cost: Decimal) -> None:
        self.quantity -= quantity
        self.cost -= cost


class TaxCalculator:
    def __init__(self, tr_hist: TrHistory, findata: FinancialData) -> None:
        self._tr_hist = tr_hist
        self._findata = findata
        self._acquisitions: dict[ISIN, list[Acquisition]] = defaultdict(list)
        self._disposals: dict[ISIN, list[Disposal]] = defaultdict(list)
        self._holdings: dict[ISIN, Section104Holding] = {}
        self._capital_gains: dict[Year, list[CapitalGain]] = defaultdict(list)

    def capital_gains(self, tax_year: Year | None = None) -> Sequence[CapitalGain]:
        self._calculate_capital_gains()

        if tax_year is not None:
            return self._capital_gains.get(tax_year, [])

        return [cg for cg_group in self._capital_gains.values() for cg in cg_group]

    def holding(self, isin: ISIN) -> Section104Holding | None:
        self._calculate_capital_gains()
        return self._holdings.get(isin)

    def get_capital_gains_table(
        self,
        tax_year: Year,
        ticker_filter: Ticker | None,
        gains_only: bool,
        losses_only: bool,
    ) -> tuple[PrettyTable, CapitalGainsSummary]:
        assert not (gains_only and losses_only)

        self._calculate_capital_gains()

        table = PrettyTable(
            field_names=(
                "Disposal Date",
                "Identification",
                "Security Name",
                "ISIN",
                "Quantity",
                "Cost (£)",
                "Proceeds (£)",
                "Gain/loss (£)",
            ),
        )

        num_disposals = 0
        disposal_proceeds = total_cost = total_gains = total_losses = Decimal("0.0")

        for cg in self.capital_gains(tax_year):
            if ticker_filter is not None and cg.disposal.ticker != ticker_filter:
                continue

            if gains_only and cg.gain_loss < 0.0:
                continue

            if losses_only and cg.gain_loss > 0.0:
                continue

            table.add_row(
                [
                    cg.disposal.date,
                    cg.identification,
                    cg.disposal.name,
                    cg.disposal.isin,
                    cg.quantity,
                    cg.cost,
                    cg.disposal.amount,
                    cg.gain_loss,
                ]
            )

            num_disposals += 1
            disposal_proceeds += round(cg.disposal.amount, 2)
            total_cost += round(cg.cost, 2)
            if cg.gain_loss > 0.0:
                total_gains += cg.gain_loss
            else:
                total_losses += abs(cg.gain_loss)

        summary = CapitalGainsSummary(
            num_disposals, disposal_proceeds, total_cost, total_gains, total_losses
        )

        return table, summary

    def get_holdings_table(
        self, ticker_filter: Ticker | None = None, show_gain_loss: bool = False
    ) -> PrettyTable:
        self._calculate_capital_gains()

        table = PrettyTable(
            field_names=(
                "Security Name",
                "ISIN",
                "Cost (£)",
                "Quantity",
                "Current Value (£)",
                "Gain/Loss (£)",
                "Weight (%)",
            ),
            hidden_fields=(
                "Current Value (£)",
                "Gain/Loss (£)",
                "Weight (%)",
            )
            if not show_gain_loss
            else (),
            show_total_fields=(
                "Current Value (£)",
                "Gain/Loss (£)",
                "Weight (%)",
            )
            if show_gain_loss
            else (),
        )

        holdings = []

        if ticker_filter is None:
            holdings = sorted(
                self._holdings.items(), key=lambda x: x[1].cost, reverse=True
            )
        else:
            try:
                isin = self._tr_hist.get_ticker_isin(ticker_filter)
            except AmbiguousTickerError as e:
                logger.warning(e)
            else:
                if isin in self._holdings:
                    holdings = [(isin, self._holdings[isin])]

        holding2value = (
            {
                isin: value
                for isin, holding in holdings
                if (value := self._get_holding_value(isin, holding)) is not None
            }
            if show_gain_loss
            else {}
        )

        portfolio_value = sum(val for val in holding2value.values())
        last_idx = len(holdings) - 1

        for idx, (isin, holding) in enumerate(holdings):
            gain_loss: Decimal | None = None
            weight: Decimal | None = None

            if holding_value := holding2value.get(isin):
                gain_loss = holding.cost - holding_value
                weight = holding_value / portfolio_value * 100

            table.add_row(
                [
                    self._tr_hist.get_security_name(isin),
                    isin,
                    holding.cost,
                    holding.quantity,
                    holding_value,
                    gain_loss or "n/a",
                    weight or "n/a",
                ],
                divider=idx == last_idx,
            )

        return table

    def disposal_years(self) -> Sequence[Year]:
        self._calculate_capital_gains()
        return list(self._capital_gains.keys())

    def _calculate_capital_gains(self) -> None:
        if self._capital_gains or self._holdings:
            # Capital gains already calculated.
            return

        logger.info("Calculating capital gains")

        # First normalise the orders by retroactively adjusting their
        # share quantity for any eventual share sub-division or share
        # consolidation event.
        orders = self._normalise_orders(self._tr_hist.orders)

        # Group together orders that have the same isin, date and type.
        same_day = self._group_same_day(orders)

        for isin, name in self._tr_hist.securities:
            logger.debug("Calculating capital gains for %s (%s)", name, isin)

            # Merge orders that were issued in the same day and have
            # the same type, and place them in the acquisitions or
            # disposals bucket.
            self._merge_same_day(isin, same_day)

            # Match disposed shares with shares acquired in the same
            # day.
            self._match_shares(isin, same_day_match)

            # Match disposed shares with shares acquired up to 30 days
            # past the disposal date.
            self._match_shares(isin, thirty_days_match)

            # Process shares disposed from a Section 104 pool.
            self._process_section104_disposals(isin)

        # Capital gains are calculated ticker by ticker in order to be
        # able to show all the intermediary calculations grouped together
        # as a future improvement. However, for the report, it is more
        # intuitive to show the capital gain transactions ordered by
        # their disposal date.
        for year, events in self._capital_gains.items():
            self._capital_gains[year] = sorted(
                events, key=lambda te: (te.disposal.timestamp, te.disposal.isin)
            )

    def _normalise_orders(self, orders: Sequence[Order]) -> Sequence[Order]:
        return [
            o.adjust_quantity(self._findata.get_security_info(o.isin).splits)
            for o in orders
        ]

    def _group_same_day(self, orders: Sequence[Order]) -> GroupDict:
        same_day = defaultdict(list)
        for o in orders:
            key = GroupKey(o.isin, o.date, type(o))
            same_day[key].append(o)
        return same_day

    def _merge_same_day(self, isin: ISIN, same_day: GroupDict) -> None:
        security_orders = (
            orders for key, orders in same_day.items() if key.isin == isin
        )

        for orders in security_orders:
            if len(orders) > 1:
                order = Order.merge(*orders)
                logger.debug('    New "same-day" merged order: %s', order)
            else:
                order = orders[0]

            if isinstance(order, Acquisition):
                self._acquisitions[isin].append(order)
            elif isinstance(order, Disposal):
                self._disposals[isin].append(order)

    def _match_shares(
        self, isin: ISIN, match_fn: Callable[[Acquisition, Disposal], bool]
    ):
        acquisits = self._acquisitions[isin]
        disposals = self._disposals[isin]
        matched: set[Order] = set()

        a_idx = 0
        d_idx = 0

        while d_idx < len(disposals):
            if a_idx == len(acquisits):
                a_idx = 0
                d_idx += 1
                continue

            a = acquisits[a_idx]
            d = disposals[d_idx]

            if not match_fn(a, d) or a in matched:
                a_idx += 1
                continue

            matched.add(a)
            matched.add(d)

            if a.quantity > d.quantity:
                a, acquisits[a_idx] = a.split(d.quantity)
                a_idx = 0
                d_idx += 1
            elif d.quantity > a.quantity:
                d, disposals[d_idx] = d.split(a.quantity)
                a_idx += 1
            else:
                a_idx = 0
                d_idx += 1

            self._capital_gains[d.tax_year()].append(
                CapitalGain(d, a.total_cost + d.fees, a.date)
            )

        self._acquisitions[isin] = [o for o in acquisits if o not in matched]
        self._disposals[isin] = [o for o in disposals if o not in matched]

    def _process_section104_disposals(self, isin: ISIN) -> None:
        security_orders = sorted(
            [*self._acquisitions[isin], *self._disposals[isin]],
            key=lambda order: order.date,
        )

        for order in security_orders:
            holding = self._holdings.get(isin)

            if isinstance(order, Acquisition):
                if holding is not None:
                    holding.increase(order.date, order.quantity, order.total_cost)
                else:
                    self._holdings[isin] = Section104Holding(
                        order.date, order.quantity, order.total_cost
                    )
            elif isinstance(order, Disposal):
                if holding is not None:
                    allowable_cost = holding.cost * order.quantity / holding.quantity

                    holding.decrease(order.date, order.quantity, allowable_cost)

                    if holding.quantity < 0.0:
                        raise_or_warn(
                            IncompleteRecordsError(
                                isin, self._tr_hist.get_security_name(isin) or "?"
                            )
                        )
                        logger.warning("Not calculating holding for %s", isin)
                        del self._holdings[isin]
                        break

                    if holding.quantity == Decimal("0.0"):
                        del self._holdings[isin]

                    self._capital_gains[order.tax_year()].append(
                        CapitalGain(order, allowable_cost + order.fees)
                    )
                else:
                    raise_or_warn(
                        IncompleteRecordsError(
                            isin, self._tr_hist.get_security_name(isin) or "?"
                        )
                    )
                    logger.warning("Not calculating holding for %s", isin)
                    break

    def _get_holding_value(
        self, isin: ISIN, holding: Section104Holding
    ) -> Decimal | None:
        if (price := self._findata.get_security_price(isin)) and (
            price_gbp := self._findata.convert_currency(price.amount, price.currency)
        ):
            return holding.quantity * price_gbp

        return None



from collections.abc import Callable, Sequence
from datetime import date, datetime, timedelta, timezone
from decimal import Decimal

import pytest
from iso4217 import Currency

from investir.config import config
from investir.exceptions import IncompleteRecordsError
from investir.findata import (
    DataProviderError,
    FinancialData,
    Price,
    SecurityInfo,
    Split,
    YahooFinanceExchangeRateProvider,
    YahooFinanceSecurityInfoProvider,
)
from investir.taxcalculator import TaxCalculator
from investir.transaction import Acquisition, Disposal, Order
from investir.trhistory import TrHistory
from investir.typing import ISIN, Ticker


@pytest.fixture(name="make_tax_calculator")
def _make_tax_calculator(mocker, tmp_path) -> Callable:
    def _method(
        orders: Sequence[Order],
        splits: Sequence[Split] | None = None,
        price: Price | Exception | None = None,
        fx_rate: Decimal | Exception | None = None,
    ) -> TaxCalculator:
        config.reset()

        if splits is None:
            splits = []
        else:
            assert all(orders[0].isin == o.isin for o in orders)

        security_info_provider = YahooFinanceSecurityInfoProvider()
        exchange_rate_provider = YahooFinanceExchangeRateProvider()
        mocker.patch.object(
            security_info_provider,
            "fech_info",
            return_value=SecurityInfo(splits=splits),
        )
        mocker.patch.object(
            security_info_provider,
            "fetch_price",
            side_effect=[price],
        )
        mocker.patch.object(
            exchange_rate_provider, "fetch_exchange_rate", side_effect=[fx_rate]
        )

        tr_hist = TrHistory(orders=orders)
        cache_file = tmp_path / "cache.yaml"

        financial_data = FinancialData(
            security_info_provider, exchange_rate_provider, tr_hist, cache_file
        )

        return TaxCalculator(tr_hist, financial_data)

    return _method


def test_section_104_disposal(make_tax_calculator):
    """
    Test Section 104 disposals using HMRC example:
      https://assets.publishing.service.gov.uk/media/65f993439316f5001164c2d7/HS284_Example_3_2024.pdf
    """
    order1 = Acquisition(
        datetime(2015, 4, 1, tzinfo=timezone.utc),
        isin=ISIN("LOBS"),
        quantity=Decimal("1000.0"),
        amount=Decimal("4000.0"),
        fees=Decimal("150.0"),
    )

    order2 = Acquisition(
        datetime(2018, 9, 1, tzinfo=timezone.utc),
        isin=ISIN("LOBS"),
        quantity=Decimal("500.0"),
        amount=Decimal("2050.0"),
        fees=Decimal("80.0"),
    )

    order3 = Disposal(
        datetime(2023, 5, 1, tzinfo=timezone.utc),
        isin=ISIN("LOBS"),
        quantity=Decimal("700.0"),
        amount=Decimal("3360.0"),
        fees=Decimal("100.0"),
    )

    order4 = Disposal(
        datetime(2024, 2, 1, tzinfo=timezone.utc),
        isin=ISIN("LOBS"),
        quantity=Decimal("400.0"),
        amount=Decimal("2080.0"),
        fees=Decimal("105.0"),
    )

    tax_calculator = make_tax_calculator([order1, order2, order3, order4])
    capital_gains = tax_calculator.capital_gains()
    assert len(capital_gains) == 2
    assert capital_gains == tax_calculator.capital_gains(2023)

    cg = capital_gains[0]
    assert cg.disposal.date == date(2023, 5, 1)
    assert cg.acquisition_date is None
    assert cg.cost.quantize(Decimal("0.00")) == Decimal("3030.67")
    assert cg.gain_loss.quantize(Decimal("0.00")) == Decimal("329.33")
    assert str(cg) == (
        "2023-05-01 LOBS quantity: 700.0, cost: £3030.67"
        ", proceeds: £3360.0, gain: £329.33"
        ", identification: Section 104"
    )

    cg = capital_gains[1]
    assert cg.disposal.date == date(2024, 2, 1)
    assert cg.acquisition_date is None
    assert cg.cost.quantize(Decimal("0.00")) == Decimal("1779.67")
    assert cg.gain_loss.quantize(Decimal("0.00")) == Decimal("300.33")

    holding = tax_calculator.holding(Ticker("LOBS"))
    assert holding.quantity == Decimal("400.0")
    assert holding.cost.quantize(Decimal("0.00")) == Decimal("1674.67")


def test_section_104_with_no_disposal_made(make_tax_calculator):
    order1 = Acquisition(
        datetime(2015, 4, 1, tzinfo=timezone.utc),
        isin=ISIN("X"),
        quantity=Decimal("1000.0"),
        amount=Decimal("4000.0"),
        fees=Decimal("150.0"),
    )

    order2 = Acquisition(
        datetime(2018, 9, 1, tzinfo=timezone.utc),
        isin=ISIN("X"),
        quantity=Decimal("500.0"),
        amount=Decimal("2000.0"),
        fees=Decimal("50.0"),
    )

    tax_calculator = make_tax_calculator([order1, order2])
    holding = tax_calculator.holding(Ticker("X"))
    assert holding.quantity == Decimal("1500.0")
    assert holding.cost == Decimal("6200.0")


def test_same_day_rule(make_tax_calculator):
    order1 = Acquisition(
        datetime(2018, 1, 1, tzinfo=timezone.utc),
        isin=ISIN("X"),
        quantity=Decimal("10.0"),
        amount=Decimal("100.0"),
    )

    order2 = Disposal(
        datetime(2019, 1, 20, 14, 0, tzinfo=timezone.utc),
        isin=ISIN("X"),
        quantity=Decimal("1.0"),
        amount=Decimal("70.0"),
    )

    order3 = Acquisition(
        datetime(2019, 1, 20, 15, 0, tzinfo=timezone.utc),
        isin=ISIN("X"),
        quantity=Decimal("1.0"),
        amount=Decimal("60.0"),
    )

    order4 = Acquisition(
        datetime(2019, 1, 20, 16, 0, tzinfo=timezone.utc),
        isin=ISIN("X"),
        quantity=Decimal("2.0"),
        amount=Decimal("65.0"),
    )

    order5 = Disposal(
        datetime(2019, 1, 20, 17, 0, tzinfo=timezone.utc),
        isin=ISIN("X"),
        quantity=Decimal("4.0"),
        amount=Decimal("280.0"),
    )

    order6 = Acquisition(
        datetime(2019, 1, 20, 18, 0, tzinfo=timezone.utc),
        isin=ISIN("X"),
        quantity=Decimal("2.0"),
        amount=Decimal("55.0"),
    )

    tax_calculator = make_tax_calculator(
        [order1, order2, order3, order4, order5, order6]
    )
    capital_gains = tax_calculator.capital_gains()
    assert len(capital_gains) == 1

    cg = capital_gains[0]
    assert cg.acquisition_date == date(2019, 1, 20)
    assert cg.disposal.amount == order2.amount + order5.amount
    assert cg.cost == order3.amount + order4.amount + order6.amount
    assert cg.gain_loss == Decimal("170.00")

    holding = tax_calculator.holding(Ticker("X"))
    assert holding.quantity == Decimal("10.0")
    assert holding.cost == Decimal("100.0")


@pytest.mark.parametrize("days_elapsed", range(1, 31))
def test_bed_and_breakfast_rule(days_elapsed, make_tax_calculator):
    order1 = Acquisition(
        datetime(2019, 1, 18, tzinfo=timezone.utc),
        isin=ISIN("X"),
        quantity=Decimal("10.0"),
        amount=Decimal("100.0"),
    )

    order2 = Disposal(
        datetime(2019, 1, 20, tzinfo=timezone.utc),
        isin=ISIN("X"),
        quantity=Decimal("5.0"),
        amount=Decimal("150.0"),
    )

    order3 = Acquisition(
        order2.timestamp + timedelta(days=days_elapsed),
        isin=ISIN("X"),
        quantity=Decimal("5.0"),
        amount=Decimal("120.0"),
    )

    tax_calculator = make_tax_calculator([order1, order2, order3])
    capital_gains = tax_calculator.capital_gains()
    assert len(capital_gains) == 1

    cg = capital_gains[0]
    assert cg.acquisition_date == order2.date + timedelta(days=days_elapsed)
    assert cg.cost == Decimal("120.0")
    assert cg.gain_loss == Decimal("30.0")
    assert str(cg) == (
        f"2019-01-20 X    quantity: 5.0, cost: £120.00"
        f", proceeds: £150.0, gain: £30.00"
        f", identification: Bed & B. ({cg.acquisition_date})"
    )

    holding = tax_calculator.holding(Ticker("X"))
    assert holding.quantity == Decimal("10")
    assert holding.cost == Decimal("100.0")


def test_acquisitions_are_not_matched_after_thirty_days_of_disposal_date(
    make_tax_calculator,
):
    order1 = Acquisition(
        datetime(2018, 1, 1, tzinfo=timezone.utc),
        isin=ISIN("X"),
        quantity=Decimal("10.0"),
        amount=Decimal("100.0"),
    )

    order2 = Disposal(
        datetime(2019, 1, 19, tzinfo=timezone.utc),
        isin=ISIN("X"),
        quantity=Decimal("5.0"),
        amount=Decimal("150.0"),
    )

    order3 = Acquisition(
        datetime(2019, 2, 19, tzinfo=timezone.utc),
        isin=ISIN("X"),
        quantity=Decimal("5.0"),
        amount=Decimal("120.0"),
    )

    tax_calculator = make_tax_calculator([order1, order2, order3])
    capital_gains = tax_calculator.capital_gains()
    assert len(capital_gains) == 1

    cg = capital_gains[0]
    assert cg.acquisition_date is None
    assert cg.cost == Decimal("50.0")
    assert cg.gain_loss == Decimal("100.0")

    holding = tax_calculator.holding(Ticker("X"))
    assert holding.quantity == Decimal("10")
    assert holding.cost == Decimal("170.0")


def test_acquisitions_are_not_matched_before_disposal_date(make_tax_calculator):
    order1 = Acquisition(
        datetime(2018, 1, 1, tzinfo=timezone.utc),
        isin=ISIN("X"),
        quantity=Decimal("10.0"),
        amount=Decimal("100.0"),
    )

    order2 = Acquisition(
        datetime(2019, 2, 18, tzinfo=timezone.utc),
        isin=ISIN("X"),
        quantity=Decimal("5.0"),
        amount=Decimal("200.0"),
    )

    order3 = Disposal(
        datetime(2019, 2, 19, tzinfo=timezone.utc),
        isin=ISIN("X"),
        quantity=Decimal("5.0"),
        amount=Decimal("150.0"),
    )

    tax_calculator = make_tax_calculator([order1, order2, order3])
    capital_gains = tax_calculator.capital_gains()
    assert len(capital_gains) == 1

    cg = capital_gains[0]
    assert cg.acquisition_date is None
    assert cg.cost == Decimal("100.00")
    assert cg.gain_loss == Decimal("50.0")

    holding = tax_calculator.holding(Ticker("X"))
    assert holding.quantity == Decimal("10")
    assert holding.cost == Decimal("200.0")


def test_same_day_rule_has_priority_to_bed_and_breakfast_rule(make_tax_calculator):
    """
    Verify that the same day rule has priority over the
    bed and breakfast rule. Note that in the list of
    tax events the bed and breakfast disposal is shown
    first due the earlier disposal date.
    """
    order1 = Acquisition(
        datetime(2018, 1, 1, tzinfo=timezone.utc),
        isin=ISIN("X"),
        quantity=Decimal("10.0"),
        amount=Decimal("100.0"),
    )

    order2 = Disposal(
        datetime(2019, 1, 20, tzinfo=timezone.utc),
        isin=ISIN("X"),
        quantity=Decimal("5.0"),
        amount=Decimal("150.0"),
    )

    order3 = Acquisition(
        datetime(2019, 1, 25, tzinfo=timezone.utc),
        isin=ISIN("X"),
        quantity=Decimal("5.0"),
        amount=Decimal("150.0"),
    )

    order4 = Disposal(
        datetime(2019, 1, 25, tzinfo=timezone.utc),
        isin=ISIN("X"),
        quantity=Decimal("5.0"),
        amount=Decimal("170.0"),
    )

    order5 = Acquisition(
        datetime(2019, 1, 27, tzinfo=timezone.utc),
        isin=ISIN("X"),
        quantity=Decimal("5.0"),
        amount=Decimal("300.0"),
    )

    tax_calculator = make_tax_calculator([order1, order2, order3, order4, order5])
    capital_gains = tax_calculator.capital_gains()
    assert len(capital_gains) == 2

    cg = capital_gains[0]  # Bed and breakfast disposal event
    assert cg.acquisition_date == date(2019, 1, 27)
    assert cg.cost == Decimal("300.0")
    assert cg.gain_loss == Decimal("-150.0")

    cg = capital_gains[1]  # Same day disposal event
    assert cg.acquisition_date == date(2019, 1, 25)
    assert cg.cost == Decimal("150.0")
    assert cg.gain_loss == Decimal("20.0")

    holding = tax_calculator.holding(Ticker("X"))
    assert holding.quantity == Decimal("10.0")
    assert holding.cost == Decimal("100.0")


def test_matching_disposals_with_larger_acquisition(make_tax_calculator):
    """
    Test splitting an acquisition in two to match the first disposal,
    and then using the remaining part to match the second disposal.
    """
    order1 = Acquisition(
        datetime(2018, 1, 1, tzinfo=timezone.utc),
        isin=ISIN("X"),
        quantity=Decimal("10.0"),
        amount=Decimal("100.0"),
    )

    order2 = Disposal(
        datetime(2019, 1, 20, tzinfo=timezone.utc),
        isin=ISIN("X"),
        quantity=Decimal("5.0"),
        amount=Decimal("60.0"),
    )

    order3 = Disposal(
        datetime(2019, 1, 21, tzinfo=timezone.utc),
        isin=ISIN("X"),
        quantity=Decimal("1.0"),
        amount=Decimal("11.0"),
    )

    order4 = Acquisition(
        datetime(2019, 1, 22, tzinfo=timezone.utc),
        isin=ISIN("X"),
        quantity=Decimal("7.0"),
        amount=Decimal("70.0"),
    )

    tax_calculator = make_tax_calculator([order1, order2, order3, order4])
    capital_gains = tax_calculator.capital_gains()
    assert len(capital_gains) == 2

    cg = capital_gains[0]
    assert cg.acquisition_date == date(2019, 1, 22)
    assert cg.cost == Decimal("50.0")
    assert cg.gain_loss == Decimal("10.0")

    cg = capital_gains[1]
    assert cg.acquisition_date == date(2019, 1, 22)
    assert cg.cost == Decimal("10.0")
    assert cg.gain_loss == Decimal("1.0")

    holding = tax_calculator.holding(Ticker("X"))
    assert holding.quantity == Decimal("11.0")
    assert holding.cost == Decimal("110.0")


def test_matching_disposal_with_multiple_smaller_acquisitions(make_tax_calculator):
    """
    Test multiple tax events derived from a single disposal
       - 1 share matched due same day rule
       - 2 shares matched due bed and breakfast rule on 2019/1/25
       - 1 shares matched due bed and breakfast rule on 2019/1/27
       - 1 share from the Section 104 pool

    In addition, verify that the second disposal does not match
    against acquisitions that were already matched for the first
    disposal.
    """
    order1 = Acquisition(
        datetime(2018, 1, 1, tzinfo=timezone.utc),
        isin=ISIN("X"),
        quantity=Decimal("10.0"),
        amount=Decimal("30.0"),
    )

    order2 = Disposal(
        datetime(2019, 1, 20, tzinfo=timezone.utc),
        isin=ISIN("X"),
        quantity=Decimal("5.0"),
        amount=Decimal("50.0"),
    )

    order3 = Acquisition(
        datetime(2019, 1, 20, tzinfo=timezone.utc),
        isin=ISIN("X"),
        quantity=Decimal("1.0"),
        amount=Decimal("9.0"),
    )

    order4 = Acquisition(
        datetime(2019, 1, 25, tzinfo=timezone.utc),
        isin=ISIN("X"),
        quantity=Decimal("2.0"),
        amount=Decimal("16.0"),
    )

    order5 = Acquisition(
        datetime(2019, 1, 27, tzinfo=timezone.utc),
        isin=ISIN("X"),
        quantity=Decimal("1.0"),
        amount=Decimal("5.0"),
    )

    order6 = Disposal(
        datetime(2019, 3, 1, tzinfo=timezone.utc),
        isin=ISIN("X"),
        quantity=Decimal("1.0"),
        amount=Decimal("7.0"),
    )

    tax_calculator = make_tax_calculator(
        [order1, order2, order3, order4, order5, order6]
    )
    capital_gains = tax_calculator.capital_gains()
    assert len(capital_gains) == 5

    cg = capital_gains[0]  # Same day
    assert cg.acquisition_date == date(2019, 1, 20)
    assert cg.cost == Decimal("9.0")
    assert cg.gain_loss == Decimal("1.0")

    cg = capital_gains[1]  # First bed and breakfast match
    assert cg.acquisition_date == date(2019, 1, 25)
    assert cg.cost == Decimal("16.0")
    assert cg.gain_loss == Decimal("4.0")

    cg = capital_gains[2]  # Second bed and breakfast match
    assert cg.acquisition_date == date(2019, 1, 27)
    assert cg.cost == Decimal("5.0")
    assert cg.gain_loss == Decimal("5.0")

    cg = capital_gains[3]  # Section 104 pool
    assert cg.acquisition_date is None
    assert cg.cost == Decimal("3.0")
    assert cg.gain_loss == Decimal("7.0")

    cg = capital_gains[4]  # Section 104 pool
    assert cg.acquisition_date is None
    assert cg.cost == Decimal("3.0")
    assert cg.gain_loss == Decimal("4.0")

    holding = tax_calculator.holding(Ticker("X"))
    assert holding.quantity == Decimal("8.0")
    assert holding.cost == Decimal("24.0")


def test_capital_gains_on_orders_with_fees_included(make_tax_calculator):
    order1 = Acquisition(
        datetime(2018, 1, 1, tzinfo=timezone.utc),
        isin=ISIN("X"),
        quantity=Decimal("10.0"),
        amount=Decimal("30.0"),
        fees=Decimal("1.5"),
    )

    order2 = Acquisition(
        datetime(2019, 1, 20, tzinfo=timezone.utc),
        isin=ISIN("X"),
        quantity=Decimal("5.0"),
        amount=Decimal("40.0"),
        fees=Decimal("0.5"),
    )

    order3 = Disposal(
        datetime(2019, 1, 20, tzinfo=timezone.utc),
        isin=ISIN("X"),
        quantity=Decimal("5.0"),
        amount=Decimal("50.0"),
        fees=Decimal("0.4"),
    )

    order4 = Disposal(
        datetime(2019, 3, 1, tzinfo=timezone.utc),
        isin=ISIN("X"),
        quantity=Decimal("5.0"),
        amount=Decimal("50.0"),
        fees=Decimal("0.8"),
    )

    tax_calculator = make_tax_calculator([order1, order2, order3, order4])
    capital_gains = tax_calculator.capital_gains()
    assert len(capital_gains) == 2

    # Acquisition cost = 40.0 + 0.4 + 0.5 = 40.9
    # Gain/Loss        = 50.0 - 40.9 = 33.45 = 9.1
    cg = capital_gains[1]
    cg = capital_gains[0]
    assert cg.acquisition_date == date(2019, 1, 20)
    assert cg.cost == Decimal("40.9")
    assert cg.gain_loss == Decimal("9.1")

    # S104 pool cost   = 30.0 + 1.5 = 31.5
    # Acquisition cost = 31.5 * (5.0/10.0) + 0.8 = 16.55
    # Gain/Loss        = 50.0 - 16.55 = 33.45
    cg = capital_gains[1]
    assert cg.acquisition_date is None
    assert cg.cost == Decimal("16.55")
    assert cg.gain_loss == Decimal("33.45")

    # New S104 pool cost = 31.5 - 31.5 * (5.0/10.0) = 15.75
    holding = tax_calculator.holding(Ticker("X"))
    assert holding.quantity == Decimal("5.0")
    assert holding.cost == Decimal("15.75")


def test_disposals_on_different_tickers(make_tax_calculator):
    order1 = Acquisition(
        datetime(2018, 1, 1, tzinfo=timezone.utc),
        isin=ISIN("X"),
        quantity=Decimal("10.0"),
        amount=Decimal("100.0"),
    )

    order2 = Acquisition(
        datetime(2018, 1, 1, tzinfo=timezone.utc),
        isin=ISIN("Y"),
        quantity=Decimal("20.0"),
        amount=Decimal("200.0"),
    )

    order3 = Disposal(
        datetime(2019, 1, 20, tzinfo=timezone.utc),
        isin=ISIN("X"),
        quantity=Decimal("1.0"),
        amount=Decimal("11.0"),
    )

    order4 = Disposal(
        datetime(2019, 1, 21, tzinfo=timezone.utc),
        isin=ISIN("Y"),
        quantity=Decimal("2.0"),
        amount=Decimal("22.0"),
    )

    order5 = Acquisition(
        datetime(2019, 1, 21, tzinfo=timezone.utc),
        isin=ISIN("X"),
        quantity=Decimal("2.0"),
        amount=Decimal("8.0"),
    )

    order6 = Acquisition(
        datetime(2019, 1, 22, tzinfo=timezone.utc),
        isin=ISIN("Y"),
        quantity=Decimal("2.0"),
        amount=Decimal("6.0"),
    )

    tax_calculator = make_tax_calculator(
        [order1, order2, order3, order4, order5, order6]
    )
    capital_gains = tax_calculator.capital_gains()
    assert len(capital_gains) == 2

    cg = capital_gains[0]
    assert cg.acquisition_date == date(2019, 1, 21)
    assert cg.disposal.isin == ISIN("X")
    assert cg.cost == Decimal("4.0")
    assert cg.gain_loss == Decimal("7.0")

    cg = capital_gains[1]
    assert cg.acquisition_date == date(2019, 1, 22)
    assert cg.disposal.isin == ISIN("Y")
    assert cg.cost == Decimal("6.0")
    assert cg.gain_loss == Decimal("16.0")

    holding = tax_calculator.holding(Ticker("X"))
    assert holding.quantity == Decimal("11.0")
    assert holding.cost == Decimal("104.0")

    holding = tax_calculator.holding(Ticker("Y"))
    assert holding.quantity == Decimal("20.0")
    assert holding.cost == Decimal("200.0")


def test_integrity_disposal_without_acquisition(make_tax_calculator):
    order = Disposal(
        datetime(2019, 1, 17, tzinfo=timezone.utc),
        isin=ISIN("X"),
        quantity=Decimal("5.0"),
        amount=Decimal("120.0"),
    )

    tax_calculator = make_tax_calculator([order])
    with pytest.raises(IncompleteRecordsError):
        tax_calculator.capital_gains()

    tax_calculator = make_tax_calculator([order])
    config.strict = False
    tax_calculator.capital_gains()


def test_integrity_disposing_more_than_quantity_acquired(make_tax_calculator):
    order1 = Acquisition(
        datetime(2019, 1, 18, tzinfo=timezone.utc),
        isin=ISIN("X"),
        quantity=Decimal("5.0"),
        amount=Decimal("100.0"),
    )

    order2 = Disposal(
        datetime(2019, 3, 17, tzinfo=timezone.utc),
        isin=ISIN("X"),
        quantity=Decimal("10.0"),
        amount=Decimal("120.0"),
    )

    tax_calculator = make_tax_calculator([order1, order2])
    with pytest.raises(IncompleteRecordsError):
        tax_calculator.capital_gains()

    tax_calculator = make_tax_calculator([order1, order2])
    config.strict = False
    tax_calculator.capital_gains()


def test_section_104_disposal_with_share_split(make_tax_calculator):
    order1 = Acquisition(
        datetime(2014, 5, 1, tzinfo=timezone.utc),
        isin=ISIN("X"),
        quantity=Decimal("11.0"),
        amount=Decimal("3300.0"),
    )

    order2 = Disposal(
        datetime(2014, 6, 1, tzinfo=timezone.utc),
        isin=ISIN("X"),
        quantity=Decimal("1.0"),
        amount=Decimal("500.0"),
    )

    order3 = Acquisition(
        datetime(2014, 8, 1, tzinfo=timezone.utc),
        isin=ISIN("X"),
        quantity=Decimal("10.0"),
        amount=Decimal("1000.0"),
    )

    order4 = Disposal(
        datetime(2014, 9, 1, tzinfo=timezone.utc),
        isin=ISIN("X"),
        quantity=Decimal("5.0"),
        amount=Decimal("1100.0"),
    )

    tax_calculator = make_tax_calculator(
        [order1, order2, order3, order4],
        [Split(datetime(2014, 7, 1, tzinfo=timezone.utc), Decimal("3.0"))],
    )
    capital_gains = tax_calculator.capital_gains()
    assert len(capital_gains) == 2

    cg = capital_gains[0]
    assert cg.disposal.date == date(2014, 6, 1)
    assert cg.acquisition_date is None
    assert cg.cost == Decimal("300.0")
    assert cg.quantity == Decimal("1.0")
    assert cg.gain_loss == Decimal("200.0")

    cg = capital_gains[1]
    assert cg.disposal.date == date(2014, 9, 1)
    assert cg.acquisition_date is None
    assert cg.cost == Decimal("500.0")
    assert cg.quantity == Decimal("5.0")
    assert cg.gain_loss == Decimal("600.0")

    holding = tax_calculator.holding(Ticker("X"))
    assert holding.quantity == Decimal("35.0")
    assert holding.cost == Decimal("3500.0")


def test_bed_and_breakfast_rule_with_share_split(make_tax_calculator):
    order1 = Acquisition(
        datetime(2019, 1, 18, tzinfo=timezone.utc),
        isin=ISIN("X"),
        quantity=Decimal("10.0"),
        amount=Decimal("100.0"),
    )

    order2 = Disposal(
        datetime(2019, 1, 20, tzinfo=timezone.utc),
        isin=ISIN("X"),
        quantity=Decimal("5.0"),
        amount=Decimal("150.0"),
    )

    order3 = Acquisition(
        datetime(2019, 1, 30, tzinfo=timezone.utc),
        isin=ISIN("X"),
        quantity=Decimal("20.0"),
        amount=Decimal("160.0"),
    )

    tax_calculator = make_tax_calculator(
        [order1, order2, order3],
        [Split(datetime(2019, 1, 25, tzinfo=timezone.utc), Decimal("3.0"))],
    )
    capital_gains = tax_calculator.capital_gains()
    assert len(capital_gains) == 1

    cg = capital_gains[0]
    assert cg.acquisition_date == order3.date
    assert cg.cost == Decimal("120.0")
    assert cg.quantity == Decimal("5.0")
    assert cg.gain_loss == Decimal("30.0")

    holding = tax_calculator.holding(Ticker("X"))
    assert holding.quantity == Decimal("35.0")
    assert holding.cost == Decimal("140.0")


def test_rppaccounts_example(make_tax_calculator):
    """
    Rawlinson Pryde & Partners Accountants calculation example:
    https://rppaccounts.co.uk/taxation-of-shares/
    """
    order1 = Acquisition(
        datetime(2019, 1, 18, tzinfo=timezone.utc),
        isin=ISIN("X"),
        quantity=Decimal("4.0"),
        amount=Decimal("12025.0"),
    )

    order2 = Disposal(
        datetime(2019, 1, 18, tzinfo=timezone.utc),
        isin=ISIN("X"),
        quantity=Decimal("3.0"),
        amount=Decimal("9100.0"),
    )

    order3 = Disposal(
        datetime(2019, 1, 19, tzinfo=timezone.utc),
        isin=ISIN("X"),
        quantity=Decimal("1.0"),
        amount=Decimal("5000.0"),
    )

    order4 = Acquisition(
        datetime(2019, 1, 22, tzinfo=timezone.utc),
        isin=ISIN("X"),
        quantity=Decimal("1.0"),
        amount=Decimal("2000.0"),
    )

    order5 = Acquisition(
        datetime(2019, 1, 23, tzinfo=timezone.utc),
        isin=ISIN("X"),
        quantity=Decimal("5.0"),
        amount=Decimal("15000.0"),
    )

    order6 = Disposal(
        datetime(2019, 1, 24, tzinfo=timezone.utc),
        isin=ISIN("X"),
        quantity=Decimal("2.0"),
        amount=Decimal("8000.0"),
    )

    order7 = Acquisition(
        datetime(2019, 1, 25, tzinfo=timezone.utc),
        isin=ISIN("X"),
        quantity=Decimal("1.0"),
        amount=Decimal("3300.0"),
    )

    order8 = Disposal(
        datetime(2019, 1, 25, tzinfo=timezone.utc),
        isin=ISIN("X"),
        quantity=Decimal("1.0"),
        amount=Decimal("3500.0"),
    )

    order9 = Acquisition(
        datetime(2019, 1, 26, tzinfo=timezone.utc),
        isin=ISIN("X"),
        quantity=Decimal("4.0"),
        amount=Decimal("17000.0"),
    )

    order10 = Disposal(
        datetime(2019, 1, 27, tzinfo=timezone.utc),
        isin=ISIN("X"),
        quantity=Decimal("8.0"),
        amount=Decimal("70000.0"),
    )

    tax_calculator = make_tax_calculator(
        [
            order1,
            order2,
            order3,
            order4,
            order5,
            order6,
            order7,
            order8,
            order9,
            order10,
        ]
    )
    capital_gains = tax_calculator.capital_gains()
    assert len(capital_gains) == 5

    cg = capital_gains[0]
    assert cg.acquisition_date == date(2019, 1, 18)
    assert cg.cost == Decimal("9018.75")
    assert cg.gain_loss == Decimal("81.25")

    cg = capital_gains[1]
    assert cg.acquisition_date == date(2019, 1, 22)
    assert cg.cost == Decimal("2000.00")
    assert cg.gain_loss == Decimal("3000.0")

    cg = capital_gains[2]
    assert cg.acquisition_date == date(2019, 1, 26)
    assert cg.cost == Decimal("8500.00")
    assert cg.gain_loss == Decimal("-500.0")

    cg = capital_gains[3]
    assert cg.acquisition_date == date(2019, 1, 25)
    assert cg.cost == Decimal("3300.0")
    assert cg.gain_loss == Decimal("200.0")

    cg = capital_gains[4]
    assert cg.acquisition_date is None
    assert cg.cost == Decimal("26506.25")
    assert cg.gain_loss == Decimal("43493.75")

    assert tax_calculator.holding(Ticker("X")) is None


# The following tests are based on the HMRC examples for calculating the
# capital gains or losses on the disposal of cryptoassets. The gain/loss
# is calculated in the same exact way as if the assets were shares from
# a company and not cryptoassets.


def test_hmrc_example_crypto22251(make_tax_calculator):
    """
    https://www.gov.uk/hmrc-internal-manuals/cryptoassets-manual/crypto22251
    """
    order1 = Acquisition(
        datetime(2019, 1, 1, tzinfo=timezone.utc),
        isin=ISIN("TOKEN-A"),
        quantity=Decimal("100.0"),
        amount=Decimal("1000.0"),
    )

    order2 = Acquisition(
        datetime(2020, 9, 18, tzinfo=timezone.utc),
        isin=ISIN("TOKEN-A"),
        quantity=Decimal("50.0"),
        amount=Decimal("125000.0"),
    )

    order3 = Disposal(
        datetime(2020, 12, 1, tzinfo=timezone.utc),
        isin=ISIN("TOKEN-A"),
        quantity=Decimal("50.0"),
        amount=Decimal("300000.0"),
    )

    tax_calculator = make_tax_calculator([order1, order2, order3])
    capital_gains = tax_calculator.capital_gains()
    assert len(capital_gains) == 1

    cg = capital_gains[0]
    assert cg.disposal.date == date(2020, 12, 1)
    assert cg.acquisition_date is None
    assert cg.cost == Decimal("42000.0")
    assert cg.gain_loss == Decimal("258000.0")

    holding = tax_calculator.holding(ISIN("TOKEN-A"))
    assert holding.quantity == Decimal("100.0")
    assert holding.cost == Decimal("84000.0")


def test_hmrc_example_crypto22252(make_tax_calculator):
    """
    https://www.gov.uk/hmrc-internal-manuals/cryptoassets-manual/crypto22252
    """
    order1 = Acquisition(
        datetime(2019, 1, 1, tzinfo=timezone.utc),
        isin=ISIN("TOKEN-B"),
        quantity=Decimal("5000.0"),
        amount=Decimal("500.0"),
    )

    order2 = Disposal(
        datetime(2020, 6, 23, 9, 0, tzinfo=timezone.utc),
        isin=ISIN("TOKEN-B"),
        quantity=Decimal("1000.0"),
        amount=Decimal("800.0"),
    )

    order3 = Acquisition(
        datetime(2020, 6, 23, 13, 0, tzinfo=timezone.utc),
        isin=ISIN("TOKEN-B"),
        quantity=Decimal("1600.0"),
        amount=Decimal("1000.0"),
    )

    order4 = Disposal(
        datetime(2020, 6, 23, 19, 0, tzinfo=timezone.utc),
        isin=ISIN("TOKEN-B"),
        quantity=Decimal("500.0"),
        amount=Decimal("600.0"),
    )

    tax_calculator = make_tax_calculator([order1, order2, order3, order4])
    capital_gains = tax_calculator.capital_gains()
    assert len(capital_gains) == 1

    cg = capital_gains[0]
    assert cg.disposal.date == date(2020, 6, 23)
    assert cg.acquisition_date == date(2020, 6, 23)
    assert cg.cost == Decimal("937.5")
    assert cg.gain_loss == Decimal("462.5")

    holding = tax_calculator.holding(ISIN("TOKEN-B"))
    assert holding.quantity == Decimal("5100.0")
    assert holding.cost == Decimal("562.5")


def test_hmrc_example_crypto22253(make_tax_calculator):
    """
    https://www.gov.uk/hmrc-internal-manuals/cryptoassets-manual/crypto22253
    """
    order1 = Acquisition(
        datetime(2019, 1, 1, tzinfo=timezone.utc),
        isin=ISIN("TOKEN-C"),
        quantity=Decimal("2000.0"),
        amount=Decimal("1000.0"),
    )

    order2 = Disposal(
        datetime(2020, 3, 31, tzinfo=timezone.utc),
        isin=ISIN("TOKEN-C"),
        quantity=Decimal("1000.0"),
        amount=Decimal("400.0"),
    )

    order3 = Disposal(
        datetime(2020, 4, 20, tzinfo=timezone.utc),
        isin=ISIN("TOKEN-C"),
        quantity=Decimal("500.0"),
        amount=Decimal("150.0"),
    )

    order4 = Acquisition(
        datetime(2020, 4, 21, tzinfo=timezone.utc),
        isin=ISIN("TOKEN-C"),
        quantity=Decimal("700.0"),
        amount=Decimal("175.0"),
    )

    order5 = Acquisition(
        datetime(2020, 4, 28, tzinfo=timezone.utc),
        isin=ISIN("TOKEN-C"),
        quantity=Decimal("500.0"),
        amount=Decimal("100.0"),
    )

    order6 = Acquisition(
        datetime(2020, 5, 1, tzinfo=timezone.utc),
        isin=ISIN("TOKEN-C"),
        quantity=Decimal("500.0"),
        amount=Decimal("150.0"),
    )

    tax_calculator = make_tax_calculator(
        [order1, order2, order3, order4, order5, order6]
    )
    capital_gains = tax_calculator.capital_gains()
    assert len(capital_gains) == 4

    cg = capital_gains[0]
    assert cg.disposal.date == date(2020, 3, 31)
    assert cg.acquisition_date == date(2020, 4, 21)
    assert cg.cost == Decimal("175.0")
    assert cg.gain_loss == Decimal("105.0")

    cg = capital_gains[1]
    assert cg.disposal.date == date(2020, 3, 31)
    assert cg.acquisition_date == date(2020, 4, 28)
    assert cg.cost == Decimal("60.0")
    assert cg.gain_loss == Decimal("60.0")

    cg = capital_gains[2]
    assert cg.disposal.date == date(2020, 4, 20)
    assert cg.acquisition_date == date(2020, 4, 28)
    assert cg.cost == Decimal("40.0")
    assert cg.gain_loss == Decimal("20.0")

    cg = capital_gains[3]
    assert cg.disposal.date == date(2020, 4, 20)
    assert cg.acquisition_date == date(2020, 5, 1)
    assert cg.cost == Decimal("90.0")
    assert cg.gain_loss == Decimal("0.0")

    holding = tax_calculator.holding(ISIN("TOKEN-C"))
    assert holding.quantity == Decimal("2200.0")
    assert holding.cost == Decimal("1060.0")


def test_hmrc_example_crypto22254(make_tax_calculator):
    """
    https://www.gov.uk/hmrc-internal-manuals/cryptoassets-manual/crypto22254
    """
    order1 = Acquisition(
        datetime(2019, 1, 1, tzinfo=timezone.utc),
        isin=ISIN("TOKEN-D"),
        quantity=Decimal("8000.0"),
        amount=Decimal("1000.0"),
    )

    order2 = Disposal(
        datetime(2020, 1, 31, tzinfo=timezone.utc),
        isin=ISIN("TOKEN-D"),
        quantity=Decimal("5000.0"),
        amount=Decimal("500.0"),
    )

    order3 = Acquisition(
        datetime(2020, 1, 31, tzinfo=timezone.utc),
        isin=ISIN("TOKEN-D"),
        quantity=Decimal("4000.0"),
        amount=Decimal("320.0"),
    )

    order4 = Acquisition(
        datetime(2020, 1, 31, tzinfo=timezone.utc),
        isin=ISIN("TOKEN-D"),
        quantity=Decimal("1000.0"),
        amount=Decimal("75.0"),
    )

    order5 = Acquisition(
        datetime(2020, 1, 31, tzinfo=timezone.utc),
        isin=ISIN("TOKEN-D"),
        quantity=Decimal("1000.0"),
        amount=Decimal("70.0"),
    )

    order6 = Disposal(
        datetime(2020, 1, 31, tzinfo=timezone.utc),
        isin=ISIN("TOKEN-D"),
        quantity=Decimal("2000.0"),
        amount=Decimal("142.0"),
    )

    order7 = Acquisition(
        datetime(2020, 1, 31, tzinfo=timezone.utc),
        isin=ISIN("TOKEN-D"),
        quantity=Decimal("500.0"),
        amount=Decimal("35.0"),
    )

    tax_calculator = make_tax_calculator(
        [order1, order2, order3, order4, order5, order6, order7]
    )
    capital_gains = tax_calculator.capital_gains()
    assert len(capital_gains) == 2

    cg = capital_gains[0]
    assert cg.disposal.date == date(2020, 1, 31)
    assert cg.acquisition_date == date(2020, 1, 31)
    assert cg.cost == Decimal("500.0")
    assert cg.gain_loss.quantize(Decimal("0.00")) == Decimal("96.14")

    cg = capital_gains[1]
    assert cg.disposal.date == date(2020, 1, 31)
    assert cg.acquisition_date is None
    assert cg.cost == Decimal("62.5")
    assert cg.gain_loss.quantize(Decimal("0.00")) == Decimal("-16.64")

    holding = tax_calculator.holding(ISIN("TOKEN-D"))
    assert holding.quantity == Decimal("7500.0")
    assert holding.cost == Decimal("937.5")


def test_hmrc_example_crypto22255(make_tax_calculator):
    """
    https://www.gov.uk/hmrc-internal-manuals/cryptoassets-manual/crypto22255
    """
    order1 = Acquisition(
        datetime(2019, 1, 1, tzinfo=timezone.utc),
        isin=ISIN("TOKEN-E"),
        quantity=Decimal("14000.0"),
        amount=Decimal("200000.0"),
    )

    order2 = Disposal(
        datetime(2020, 8, 30, tzinfo=timezone.utc),
        isin=ISIN("TOKEN-E"),
        quantity=Decimal("4000.0"),
        amount=Decimal("160000.0"),
    )

    order3 = Acquisition(
        datetime(2020, 9, 11, tzinfo=timezone.utc),
        isin=ISIN("TOKEN-E"),
        quantity=Decimal("500.0"),
        amount=Decimal("17500.0"),
    )

    tax_calculator = make_tax_calculator([order1, order2, order3])
    capital_gains = tax_calculator.capital_gains()
    assert len(capital_gains) == 2

    cg = capital_gains[0]
    assert cg.disposal.date == date(2020, 8, 30)
    assert cg.acquisition_date == date(2020, 9, 11)
    assert cg.cost == Decimal("17500.0")
    assert cg.gain_loss == Decimal("2500.0")

    cg = capital_gains[1]
    assert cg.disposal.date == date(2020, 8, 30)
    assert cg.acquisition_date is None
    assert cg.cost == Decimal("50000.0")
    assert cg.gain_loss == Decimal("90000.0")

    holding = tax_calculator.holding(ISIN("TOKEN-E"))
    assert holding.quantity == Decimal("10500.0")
    assert holding.cost == Decimal("150000.0")


def test_hmrc_example_crypto22256(make_tax_calculator):
    """
    https://www.gov.uk/hmrc-internal-manuals/cryptoassets-manual/crypto22256
    """
    order1 = Acquisition(
        datetime(2019, 1, 1, tzinfo=timezone.utc),
        isin=ISIN("TOKEN-F"),
        quantity=Decimal("100000.0"),
        amount=Decimal("300000.0"),
    )

    order2 = Acquisition(
        datetime(2020, 7, 31, tzinfo=timezone.utc),
        isin=ISIN("TOKEN-F"),
        quantity=Decimal("10000.0"),
        amount=Decimal("45000.0"),
    )

    order3 = Disposal(
        datetime(2020, 7, 31, tzinfo=timezone.utc),
        isin=ISIN("TOKEN-F"),
        quantity=Decimal("30000.0"),
        amount=Decimal("150000.0"),
    )

    order4 = Disposal(
        datetime(2020, 8, 5, tzinfo=timezone.utc),
        isin=ISIN("TOKEN-F"),
        quantity=Decimal("20000.0"),
        amount=Decimal("100000.0"),
    )

    order5 = Acquisition(
        datetime(2020, 8, 6, tzinfo=timezone.utc),
        isin=ISIN("TOKEN-F"),
        quantity=Decimal("50000.0"),
        amount=Decimal("225000.0"),
    )

    order6 = Disposal(
        datetime(2020, 8, 7, tzinfo=timezone.utc),
        isin=ISIN("TOKEN-F"),
        quantity=Decimal("100000.0"),
        amount=Decimal("150000.0"),
    )

    tax_calculator = make_tax_calculator(
        [order1, order2, order3, order4, order5, order6]
    )
    capital_gains = tax_calculator.capital_gains()
    assert len(capital_gains) == 4

    cg = capital_gains[0]
    assert cg.disposal.date == date(2020, 7, 31)
    assert cg.acquisition_date == date(2020, 7, 31)
    assert cg.cost == Decimal("45000.0")
    assert cg.gain_loss == Decimal("5000.0")

    cg = capital_gains[1]
    assert cg.disposal.date == date(2020, 7, 31)
    assert cg.acquisition_date == date(2020, 8, 6)
    assert cg.cost == Decimal("90000.0")
    assert cg.gain_loss == Decimal("10000.0")

    cg = capital_gains[2]
    assert cg.disposal.date == date(2020, 8, 5)
    assert cg.acquisition_date == date(2020, 8, 6)
    assert cg.cost == Decimal("90000.0")
    assert cg.gain_loss == Decimal("10000.0")

    cg = capital_gains[3]
    assert cg.disposal.date == date(2020, 8, 7)
    assert cg.acquisition_date is None
    assert cg.cost.quantize(Decimal("0.00")) == Decimal("313636.36")
    assert cg.gain_loss.quantize(Decimal("0.00")) == Decimal("-163636.36")

    holding = tax_calculator.holding(ISIN("TOKEN-F"))
    assert holding.quantity == Decimal("10000.0")
    assert holding.cost.quantize(Decimal("0.00")) == Decimal("31363.64")


def test_get_holdings_table_with_gain_loss(make_tax_calculator):
    order = Acquisition(
        datetime(2019, 1, 18, tzinfo=timezone.utc),
        isin=ISIN("X"),
        ticker="TICKER",
        quantity=Decimal("10.0"),
        amount=Decimal("100.0"),
    )

    tax_calculator = make_tax_calculator(
        [order], price=Price(Decimal("15.0"), Currency.GBP)
    )
    table = tax_calculator.get_holdings_table(show_gain_loss=True)

    assert "Gain/Loss (£)" in str(table)
    assert "50.00" in str(table)


def test_get_holdings_table_with_gain_loss_and_currency_conversion(make_tax_calculator):
    order = Acquisition(
        datetime(2019, 1, 18, tzinfo=timezone.utc),
        isin=ISIN("X"),
        ticker="TICKER",
        quantity=Decimal("10.0"),
        amount=Decimal("100.0"),
    )

    tax_calculator = make_tax_calculator(
        [order],
        price=Price(Decimal("15.0"), Currency.USD),
        fx_rate=Decimal("0.75"),
    )
    table = tax_calculator.get_holdings_table(show_gain_loss=True)

    assert "Gain/Loss (£)" in str(table)
    assert "12.50" in str(table)


def test_get_holdings_table_with_gain_loss_when_price_or_fx_rate_not_available(
    make_tax_calculator,
):
    order = Acquisition(
        datetime(2019, 1, 18, tzinfo=timezone.utc),
        isin=ISIN("X"),
        ticker="TICKER",
        quantity=Decimal("10.0"),
        amount=Decimal("100.0"),
    )

    tax_calculator = make_tax_calculator([order], price=DataProviderError)
    table = tax_calculator.get_holdings_table(show_gain_loss=True)

    assert "Gain/Loss (£)" in str(table)
    assert "n/a" in str(table)

    tax_calculator = make_tax_calculator(
        [order],
        price=Price(Decimal("15.0"), Currency.USD),
        fx_rate=DataProviderError,
    )

    table = tax_calculator.get_holdings_table(show_gain_loss=True)
    assert "Gain/Loss (£)" in str(table)
    assert "n/a" in str(table)


def test_get_holdings_table_ambiguous_ticker(make_tax_calculator, capsys):
    order1 = Acquisition(
        datetime(2019, 1, 18, tzinfo=timezone.utc),
        isin=ISIN("X"),
        ticker="TICKER",
        quantity=Decimal("10.0"),
        amount=Decimal("100.0"),
    )

    order2 = Acquisition(
        datetime(2019, 1, 20, tzinfo=timezone.utc),
        isin=ISIN("Y"),
        ticker="TICKER",
        quantity=Decimal("5.0"),
        amount=Decimal("150.0"),
    )

    tax_calculator = make_tax_calculator([order1, order2])
    tax_calculator.get_holdings_table(ticker_filter=Ticker("TICKER"))

    captured = capsys.readouterr()
    assert not captured.out
    assert not captured.err



import investir.parser.freetrade as _freetrade  # noqa: F401
import investir.parser.trading212 as _trading212  # noqa: F401
from investir.parser.factory import ParserFactory

__all__ = ["ParserFactory"]



from datetime import date, datetime, timezone
from decimal import Decimal

import pytest

from investir.findata import Split
from investir.transaction import Acquisition, Disposal, Order
from investir.typing import ISIN, Ticker


def test_acquisition_order():
    count = Order.order_count

    order = Acquisition(
        datetime(2022, 4, 6, 18, 4, 50),
        isin=ISIN("AMZN-ISIN"),
        ticker=Ticker("AMZN"),
        name="Amazon",
        amount=Decimal("100.0"),
        quantity=Decimal("20.0"),
        fees=Decimal("1.4"),
        tr_id="ORDER",
    )

    assert order.date == date(2022, 4, 6)
    assert order.tax_year() == 2022
    assert order.number == count + 1
    assert order.price == order.amount / order.quantity
    assert order.total_cost == order.amount + order.fees


def test_disposal_order():
    count = Order.order_count

    order = Disposal(
        datetime(2023, 4, 6, 18, 4, 50),
        isin=ISIN("AMZN-ISIN"),
        ticker=Ticker("AMZN"),
        name="Amazon",
        amount=Decimal("50.0"),
        quantity=Decimal("10.0"),
        fees=Decimal("1.7"),
        tr_id="ORDER",
    )

    assert order.date == date(2023, 4, 6)
    assert order.tax_year() == 2023
    assert order.number == count + 1
    assert order.price == order.amount / order.quantity
    assert order.net_proceeds == order.amount - order.fees


def test_order_merge():
    order1 = Acquisition(
        datetime(2022, 4, 6, 18, 4, 50),
        isin=ISIN("AMZN-ISIN"),
        ticker=Ticker("AMZN"),
        name="Amazon",
        amount=Decimal("100.0"),
        quantity=Decimal("20.0"),
        fees=Decimal("1.4"),
    )

    order2 = Acquisition(
        datetime(2022, 4, 6, 18, 4, 50),
        isin=ISIN("AMZN-ISIN"),
        ticker=Ticker("AMZN"),
        name="Amazon",
        amount=Decimal("50.0"),
        quantity=Decimal("5.0"),
        fees=Decimal("1.5"),
    )

    order3 = Acquisition(
        datetime(2022, 4, 6, 18, 4, 50),
        isin=ISIN("AMZN-ISIN"),
        ticker=Ticker("AMZN"),
        name="Amazon",
        amount=Decimal("10.0"),
        quantity=Decimal("4.0"),
        fees=Decimal("0.1"),
    )

    merged_order = Order.merge(order1, order2, order3)
    assert merged_order.timestamp == datetime(2022, 4, 6, 0, 0)
    assert merged_order.isin == ISIN("AMZN-ISIN")
    assert merged_order.ticker == Ticker("AMZN")
    assert merged_order.name == "Amazon"
    assert merged_order.amount == Decimal("160")
    assert merged_order.quantity == Decimal("29")
    assert merged_order.fees == Decimal("3.0")


def test_order_split():
    date_time = datetime(2022, 4, 6, 18, 4, 50)

    order = Acquisition(
        date_time,
        isin=ISIN("AMZN-ISIN"),
        ticker=Ticker("AMZN"),
        name="Amazon",
        amount=Decimal("120.0"),
        quantity=Decimal("12.0"),
        fees=Decimal("6.0"),
    )

    matched, remainder = order.split(Decimal("8.0"))

    assert isinstance(matched, Acquisition)
    assert matched.timestamp == date_time
    assert matched.isin == ISIN("AMZN-ISIN")
    assert matched.ticker == Ticker("AMZN")
    assert matched.name == "Amazon"
    assert matched.amount == Decimal("80.0")
    assert matched.quantity == Decimal("8.0")
    assert matched.fees == Decimal("4.0")

    assert isinstance(remainder, Acquisition)
    assert remainder.timestamp == date_time
    assert remainder.isin == ISIN("AMZN-ISIN")
    assert remainder.ticker == Ticker("AMZN")
    assert remainder.name == "Amazon"
    assert remainder.amount == Decimal("40.0")
    assert remainder.quantity == Decimal("4.0")
    assert remainder.fees == Decimal("2.0")

    with pytest.raises(AssertionError):
        order.split(Decimal("12.1"))


def test_order_adjust_quantity():
    order1 = Acquisition(
        datetime(2018, 1, 1, tzinfo=timezone.utc),
        isin=ISIN("AMZN-ISIN"),
        name="Amazon",
        amount=Decimal("10.0"),
        quantity=Decimal("1.0"),
    )

    order2 = Acquisition(
        datetime(2020, 1, 1, tzinfo=timezone.utc),
        isin=ISIN("AMZN-ISIN"),
        name="Amazon",
        amount=Decimal("10.0"),
        quantity=Decimal("1.0"),
    )

    splits = [
        Split(datetime(2019, 1, 1, tzinfo=timezone.utc), Decimal("10.0")),
        Split(datetime(2021, 1, 1, tzinfo=timezone.utc), Decimal("3.0")),
    ]

    order2_adjusted = order2.adjust_quantity(splits)
    ratio = Decimal("3.0")
    assert type(order2_adjusted) is type(order2)
    assert order2_adjusted.timestamp == order2.timestamp
    assert order2_adjusted.isin == order2.isin
    assert order2_adjusted.ticker == order2.ticker
    assert order2_adjusted.name == order2.name
    assert order2_adjusted.amount == order2.amount
    assert order2_adjusted.fees == order2.fees
    assert order2_adjusted.quantity == order2.quantity * ratio
    assert order2_adjusted.original_quantity == order2.quantity
    assert order2_adjusted.tr_id == order2.tr_id
    assert "Adjusted from order" in order2_adjusted.notes

    order1_adjusted = order1.adjust_quantity(splits)
    ratio = Decimal("10.0") * Decimal("3.0")
    assert type(order1_adjusted) is type(order1)
    assert order1_adjusted.timestamp == order1.timestamp
    assert order1_adjusted.isin == order1.isin
    assert order1_adjusted.ticker == order1.ticker
    assert order1_adjusted.name == order1.name
    assert order1_adjusted.amount == order1.amount
    assert order1_adjusted.fees == order1.fees
    assert order1_adjusted.quantity == order1.quantity * ratio
    assert order1_adjusted.original_quantity == order1.quantity
    assert order1_adjusted.tr_id == order1.tr_id
    assert "Adjusted from order" in order1_adjusted.notes



import logging
from decimal import Decimal
from typing import Protocol

import yfinance
from iso4217 import Currency

from investir.findata.types import Price, SecurityInfo, Split
from investir.typing import ISIN

logger = logging.getLogger(__name__)


class DataProviderError(Exception):
    pass


class SecurityInfoProvider(Protocol):
    def fech_info(self, isin: ISIN) -> SecurityInfo:
        pass

    def fetch_price(self, isin: ISIN) -> Price:
        pass


class ExchangeRateProvider(Protocol):
    def fetch_exchange_rate(
        self, currency_from: Currency, currency_to: Currency
    ) -> Decimal:
        pass


class YahooFinanceSecurityInfoProvider:
    def fech_info(self, isin: ISIN) -> SecurityInfo:
        try:
            yf_data = yfinance.Ticker(isin)
            name = yf_data.info["shortName"]
        except Exception as ex:
            logger.debug("Exception from yfinance: %s", repr(ex))
            raise DataProviderError(f"Failed to fetch information for {isin}") from None

        splits = [
            Split(pd_date.to_pydatetime(), Decimal(ratio))
            for pd_date, ratio in yf_data.splits.items()
        ]

        return SecurityInfo(name, splits)

    def fetch_price(self, isin: ISIN) -> Price:
        try:
            yf_data = yfinance.Ticker(isin)
            price = Decimal(yf_data.info["currentPrice"])
            currency = yf_data.info["currency"]
        except Exception as e:
            logger.debug("Exception from yfinance: %s", repr(e))
            raise DataProviderError(
                f"Failed currency_to fetch last price for {isin}"
            ) from None

        if currency == "GBp":
            currency = "GBP"
            price *= Decimal("0.01")

        return Price(price, Currency(currency))


class YahooFinanceExchangeRateProvider(ExchangeRateProvider):
    def fetch_exchange_rate(
        self, currency_from: Currency, currency_to: Currency
    ) -> Decimal:
        try:
            yf_data = yfinance.Ticker(f"{currency_from.code}{currency_to.code}=X")
            fx_rate = Decimal(yf_data.info["bid"])
        except Exception as e:
            logger.debug("Exception from yfinance: %s", repr(e))
            raise DataProviderError(
                f"Failed to fetch exchange rate for "
                f"{currency_from.currency_name} ({currency_from.code}) to "
                f"{currency_to.currency_name} ({currency_to.code})"
            ) from None

        return fx_rate



from datetime import datetime, timezone
from typing import Final

# Minimum timestamp for an order. Different rules apply on orders made
# before 6 April 2008. See:
# https://www.gov.uk/hmrc-internal-manuals/capital-gains-manual/cg51570
MIN_TIMESTAMP: Final = datetime(2008, 4, 6, tzinfo=timezone.utc)



import operator
import sys
from abc import ABC
from collections.abc import Sequence
from dataclasses import dataclass, field, replace
from datetime import date, datetime
from decimal import Decimal
from functools import reduce
from typing import ClassVar, TypeVar

from investir.findata.types import Split
from investir.typing import ISIN, Ticker, Year
from investir.utils import date_to_tax_year

if sys.version_info >= (3, 11):
    from typing import Self
else:
    Self = TypeVar("Self", bound="Order")


@dataclass(frozen=True)
class Transaction(ABC):
    timestamp: datetime
    amount: Decimal
    tr_id: str | None = None
    notes: str | None = None

    @property
    def date(self) -> date:
        return self.timestamp.date()

    def tax_year(self) -> Year:
        return date_to_tax_year(self.date)


@dataclass(kw_only=True, frozen=True)
class Order(Transaction, ABC):
    number: int = field(default=0, compare=False)
    isin: ISIN
    ticker: Ticker | None = None
    name: str = ""
    quantity: Decimal
    original_quantity: Decimal | None = None
    fees: Decimal = Decimal("0.0")

    order_count: ClassVar[int] = 0

    def __post_init__(self) -> None:
        Order.order_count += 1
        object.__setattr__(self, "number", Order.order_count)

    @property
    def price(self) -> Decimal:
        return self.amount / self.quantity

    def split(self: Self, split_quantity: Decimal) -> tuple[Self, Self]:
        assert self.quantity >= split_quantity

        match_amount = self.price * split_quantity
        match_quantity = split_quantity
        match_fees = self.fees / self.quantity * split_quantity

        remainder_amount = self.amount - match_amount
        remainder_quantity = self.quantity - match_quantity
        remainder_fees = self.fees - match_fees

        match = replace(
            self,
            amount=match_amount,
            quantity=match_quantity,
            fees=match_fees,
            notes=f"Splitted from order {self.number}",
        )

        remainder = replace(
            self,
            amount=remainder_amount,
            quantity=remainder_quantity,
            fees=remainder_fees,
            notes=f"Splitted from order {self.number}",
        )

        return match, remainder

    @staticmethod
    def merge(*orders: "Order") -> "Order":
        assert len(orders) > 1

        isin = orders[0].isin
        assert all(order.isin == isin for order in orders)

        timestamp = orders[0].timestamp.replace(
            hour=0, minute=0, second=0, microsecond=0
        )

        amount = Decimal(sum(order.amount for order in orders))
        quantity = Decimal(sum(order.quantity for order in orders))
        fees = Decimal(sum(order.fees for order in orders))

        notes = "Merged from orders "
        notes += ",".join(str(order.number) for order in orders)

        return replace(
            orders[0],
            timestamp=timestamp,
            amount=amount,
            quantity=quantity,
            fees=fees,
            notes=notes,
        )

    def adjust_quantity(self, splits: Sequence[Split]) -> "Order":
        split_ratios = [s.ratio for s in splits if self.timestamp < s.date_effective]

        if not split_ratios:
            return self

        quantity = reduce(operator.mul, [self.quantity, *split_ratios])

        return replace(
            self,
            isin=self.isin,
            quantity=quantity,
            original_quantity=self.quantity,
            notes=(
                f"Adjusted from order {self.number} after applying the "
                f"following split ratios: {', '.join(map(str, split_ratios))}"
            ),
        )


@dataclass(frozen=True)
class Acquisition(Order):
    @property
    def total_cost(self) -> Decimal:
        return self.amount + self.fees


@dataclass(frozen=True)
class Disposal(Order):
    @property
    def net_proceeds(self) -> Decimal:
        return self.amount - self.fees


@dataclass(kw_only=True, frozen=True)
class Dividend(Transaction):
    isin: ISIN
    name: str = ""
    ticker: Ticker | None = None
    withheld: Decimal | None


@dataclass(frozen=True)
class Transfer(Transaction):
    pass


@dataclass(frozen=True)
class Interest(Transaction):
    pass



from collections.abc import Callable, Sequence, Set
from datetime import date
from decimal import Decimal
from enum import Enum
from typing import Any

import prettytable

from investir.utils import boldify, unboldify


class OutputFormat(str, Enum):
    TEXT = "text"
    CSV = "csv"
    JSON = "json"
    HTML = "html"


def date_format(format: str) -> Callable[[str, Any], str]:
    def _date_format(_field, val) -> str:
        if isinstance(val, date):
            return val.strftime(format)
        return val

    return _date_format


def decimal_format(precision: int) -> Callable[[str, Any], str]:
    def _decimal_format(_field, val) -> str:
        if isinstance(val, Decimal):
            return f"{val:.{precision}f}"
        elif isinstance(val, str):
            return val
        else:
            return ""

    return _decimal_format


class PrettyTable(prettytable.PrettyTable):
    def __init__(
        self,
        field_names: Sequence[str],
        hidden_fields: Sequence[str] | None = None,
        show_total_fields: Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(field_names, **kwargs)

        self.hrules = prettytable.HEADER
        self.vrules = prettytable.NONE
        self._hidden_fields: Set[str] = frozenset(hidden_fields or [])
        self._show_total_fields: Set[str] = frozenset(show_total_fields or [])

    def __bool__(self) -> bool:
        return len(self.rows) > 0

    def to_string(self, format: OutputFormat, leading_nl: bool = True) -> str:
        if (
            self.rows
            and self._show_total_fields
            and format in (OutputFormat.TEXT, OutputFormat.HTML)
        ):
            self._insert_totals_row()

        self._apply_formatting(bold_titles=format == OutputFormat.TEXT)

        start_nl = "\n" if leading_nl else ""
        end_nl = "\n" if format == OutputFormat.TEXT else ""

        fields = [
            f for f in self.field_names if unboldify(f) not in self._hidden_fields
        ]

        kwargs: dict[str, Any] = {"fields": fields}
        if format == OutputFormat.JSON:
            kwargs["default"] = str

        table_str = self.get_formatted_string(format, **kwargs)

        if format == OutputFormat.CSV:
            table_str = table_str.rstrip()

        return f"{start_nl}{table_str}{end_nl}"

    def _insert_totals_row(self) -> None:
        totals_row = []

        for i, f in enumerate(self.field_names):
            if unboldify(f) in self._show_total_fields:
                total = sum(
                    (
                        round(row[i], 2)
                        for row in self.rows
                        if row[i] and row[i] != "n/a"
                    ),
                    Decimal("0.0"),
                )
                totals_row.append(total)
            else:
                totals_row.append("")

        self.add_row(totals_row)

    def _apply_formatting(self, bold_titles: bool) -> None:
        if bold_titles:
            self.field_names = list(map(lambda f: boldify(f), self.field_names))

        for f in self.field_names:
            plain_f = unboldify(f) if bold_titles else f

            match plain_f.split()[0].strip(), plain_f.split()[-1]:
                case ("Date", _) | (_, "Date"):
                    self.custom_format[f] = date_format("%d/%m/%Y")
                    self.align[f] = "l"
                case ("Quantity", _):
                    self.custom_format[f] = decimal_format(8)
                    self.align[f] = "r"

                case (_, "(£)") | (_, "(%)"):
                    self.custom_format[f] = decimal_format(2)
                    self.align[f] = "r"

                case _:
                    self.align[f] = "l"



from collections.abc import Sequence
from dataclasses import dataclass, field
from datetime import datetime, timezone
from decimal import Decimal
from typing import NamedTuple

import yaml
from dateutil.parser import parse as parse_timestamp
from iso4217 import Currency


def utcnow() -> datetime:
    return datetime.now(timezone.utc).replace(microsecond=0)


class Price(NamedTuple):
    amount: Decimal
    currency: Currency


@dataclass
class Split(yaml.YAMLObject):
    date_effective: datetime
    ratio: Decimal
    yaml_tag = "!split"

    @classmethod
    def from_yaml(cls, loader, node) -> "Split":
        value = loader.construct_scalar(node)
        timestamp, ratio = value.split(",")
        return Split(parse_timestamp(timestamp), Decimal(ratio))

    @classmethod
    def to_yaml(cls, dumper, data):
        return dumper.represent_scalar(
            cls.yaml_tag, f"{data.date_effective}, {data.ratio}"
        )

    def __eq__(self, other) -> bool:
        return self.date_effective == other.date_effective and self.ratio == other.ratio


@dataclass
class SecurityInfo(yaml.YAMLObject):
    name: str = ""
    splits: Sequence[Split] = field(default_factory=list)
    last_updated: datetime = field(default_factory=utcnow)
    yaml_tag = "!security"



import csv
import importlib
import os
import re
import shlex
import sys
from collections.abc import Callable, Sequence
from pathlib import Path

import pytest
from typer.testing import CliRunner, Result

from investir.cli import app
from investir.config import config

PROJECT_DIR = Path(__file__).parent.parent
DATA_FILE = str(PROJECT_DIR / "data" / "freetrade.csv")
TEST_DIR = PROJECT_DIR / "tests" / "test_cli"
EX_OK = getattr(os, "EX_OK", 0)


@pytest.fixture(name="execute")
def fixture_execute() -> Callable:
    def _execute(
        args: Sequence[str],
        global_opts: Sequence[str] = [
            "--quiet",
            "--offline",
            "--cache-file",
            os.devnull,
        ],
    ) -> Result:
        config.reset()
        runner = CliRunner(mix_stderr=False)
        return runner.invoke(app, [*global_opts, *args])

    return _execute


test_data = [
    # Orders
    ("orders", "orders"),
    ("orders --tax-year 2022", "orders_ty2022"),
    ("orders --ticker SWKS", "orders_swks"),
    ("orders --acquisitions", "orders_acquisitions"),
    ("orders --disposals", "orders_disposals"),
    (
        "orders --tax-year 2022 --ticker SWKS",
        "orders_ty2022_swks",
    ),
    (
        "orders --tax-year 2022 --acquisitions",
        "orders_ty2022_acquisitions",
    ),
    (
        "orders --tax-year 2022 --disposals",
        "orders_ty2022_disposals",
    ),
    (
        "orders --ticker SWKS --acquisitions",
        "orders_swks_acquisitions",
    ),
    (
        "orders --ticker SWKS --disposals",
        "orders_swks_disposals",
    ),
    (
        "orders --tax-year 2022 --ticker SWKS --acquisitions",
        "orders_ty2022_swks_acquisitions",
    ),
    (
        "orders --tax-year 2022 --ticker SWKS --disposals",
        "orders_ty2022_swks_disposals",
    ),
    # Dividends
    ("dividends", "dividends"),
    ("dividends --tax-year 2022", "dividends_ty2022"),
    ("dividends --ticker AAPL", "dividends_aapl"),
    ("dividends --tax-year 2022 --ticker AAPL", "dividends_ty2022_aapl"),
    # Interest
    ("interest", "interest"),
    ("interest --tax-year 2022", "interest_ty2022"),
    # Transfers
    ("transfers", "transfers"),
    ("transfers --tax-year 2022", "transfers_ty2022"),
    ("transfers --deposits", "transfers_deposits"),
    ("transfers --withdrawals", "transfers_withdrawals"),
    ("transfers --tax-year 2022 --deposits", "transfers_ty2022_deposits"),
    ("transfers --tax-year 2022 --withdrawals", "transfers_ty2022_withdrawals"),
    # Capital gains
    ("capital-gains", "capital_gains"),
    ("capital-gains --tax-year 2022", "capital_gains_ty2022"),
    ("capital-gains --ticker SWKS", "capital_gains_swks"),
    ("capital-gains --ticker AMZN", "capital_gains_amzn"),
    ("capital-gains --gains", "capital_gains_gains"),
    ("capital-gains --losses", "capital_gains_losses"),
    ("capital-gains --ticker SWKS --gains", "capital_gains_swks_gains"),
    ("capital-gains --ticker SWKS --losses", "capital_gains_swks_losses"),
    ("capital-gains --tax-year 2022 --gains", "capital_gains_ty2022_gains"),
    ("capital-gains --tax-year 2022 --losses", "capital_gains_ty2022_losses"),
    (
        "capital-gains --tax-year 2022 --ticker SWKS --gains",
        "capital_gains_ty2022_swks_gains",
    ),
    (
        "capital-gains --tax-year 2022 --ticker SWKS --losses",
        "capital_gains_ty2022_swks_losses",
    ),
    # Holdings
    ("holdings", "holdings"),
    ("holdings --ticker SWKS", "holdings_swks"),
    # Output formats
    ("orders --output text", "orders"),
    ("orders --output csv", "orders_csv"),
    ("orders --output json", "orders_json"),
    ("orders --output html", "orders_html"),
    ("dividends --output text", "dividends"),
    ("dividends --output csv", "dividends_csv"),
    ("dividends --output json", "dividends_json"),
    ("dividends --output html", "dividends_html"),
    ("interest --output text", "interest"),
    ("interest --output csv", "interest_csv"),
    ("interest --output json", "interest_json"),
    ("interest --output html", "interest_html"),
    ("transfers --output text", "transfers"),
    ("transfers --output csv", "transfers_csv"),
    ("transfers --output json", "transfers_json"),
    ("transfers --output html", "transfers_html"),
    ("capital-gains --output text", "capital_gains"),
    ("capital-gains --tax-year 2022 --output csv", "capital_gains_csv"),
    ("capital-gains --tax-year 2022 --output json", "capital_gains_json"),
    ("capital-gains --tax-year 2022 --output html", "capital_gains_html"),
    ("holdings --output text", "holdings"),
    ("holdings --output csv", "holdings_csv"),
    ("holdings --output json", "holdings_json"),
    ("holdings --output html", "holdings_html"),
]


@pytest.mark.parametrize("cmd,expected", test_data)
def test_cli_output(request, execute, cmd, expected):
    expected_output = TEST_DIR / expected
    result = execute([*shlex.split(cmd), DATA_FILE])

    if request.config.getoption("--regen-outputs"):
        expected_output.write_text(result.stdout)
    else:
        assert not result.stderr
        assert result.exit_code == EX_OK

        assert expected_output.exists()
        assert (
            result.stdout.splitlines()
            == expected_output.read_text(encoding="utf-8").splitlines()
        )


def test_capital_gains_multiple_input_files(execute):
    expected_output = TEST_DIR / "capital_gains"
    data_file1 = str(PROJECT_DIR / "data" / "trading212_2021-2022.csv")
    data_file2 = str(PROJECT_DIR / "data" / "trading212_2022-2023.csv")
    result = execute(["capital-gains", data_file1, data_file2])

    assert not result.stderr
    assert result.exit_code == EX_OK

    assert expected_output.exists()
    assert result.stdout == expected_output.read_text(encoding="utf-8")


def test_orders_command_no_results(execute):
    result = execute(["orders", "--tax-year", "2008", DATA_FILE])
    assert not result.stdout
    assert not result.stderr
    assert result.exit_code == EX_OK


def test_dividends_command_no_results(execute):
    result = execute(["dividends", "--tax-year", "2008", DATA_FILE])
    assert not result.stdout
    assert not result.stderr
    assert result.exit_code == EX_OK


def test_transfers_command_no_results(execute):
    result = execute(["transfers", "--tax-year", "2008", DATA_FILE])
    assert not result.stdout
    assert not result.stderr
    assert result.exit_code == EX_OK


def test_interest_command_no_results(execute):
    result = execute(["interest", "--tax-year", "2008", DATA_FILE])
    assert not result.stdout
    assert not result.stderr
    assert result.exit_code == EX_OK


def test_capital_gains_command_no_results(execute):
    result = execute(["capital-gains", "--tax-year", "2008", DATA_FILE])
    assert not result.stdout
    assert not result.stderr
    assert result.exit_code == EX_OK


def test_holdings_no_results(execute):
    result = execute(["holdings", "--ticker", "NOTFOUND", DATA_FILE])
    assert not result.stdout
    assert not result.stderr
    assert result.exit_code == EX_OK


def test_verbosity_mutually_exclusive_filters(execute):
    result = execute(["--quiet", "--verbose", "orders", DATA_FILE])
    assert not result.stdout
    assert "Usage:" in result.stderr
    assert result.exit_code != EX_OK


def test_orders_command_mutually_exclusive_filters(execute):
    result = execute(["orders", "--acquisitions", "--disposals", DATA_FILE])
    assert not result.stdout
    assert "Usage:" in result.stderr
    assert result.exit_code != EX_OK


def test_transfers_command_mutually_exclusive_filters(execute):
    result = execute(["transfers", "--deposits", "--withdrawals", DATA_FILE])
    assert not result.stdout
    assert "Usage:" in result.stderr
    assert result.exit_code != EX_OK


def test_capital_gains_command_mutually_exclusive_filters(execute):
    result = execute(["capital-gains", "--gains", "--losses", DATA_FILE])
    assert not result.stdout
    assert "Usage:" in result.stderr
    assert result.exit_code != EX_OK


def test_capital_gains_command_tax_year_required(execute):
    result = execute(["capital-gains", "--output", "json", DATA_FILE])
    assert not result.stdout
    assert "Usage:" in result.stderr
    assert result.exit_code != EX_OK


def test_invocation_without_any_argument(execute):
    result = execute([], global_opts=[])
    assert not result.stderr
    assert "Options" in result.stdout
    assert "Commands" in result.stdout
    assert re.search(
        r"orders.*dividends.*transfers.*interest.*capital-gains.*holdings",
        result.stdout,
        re.DOTALL,
    )
    assert result.exit_code == EX_OK


def test_version_option(execute):
    result = execute(["--version"], global_opts=[])
    version = importlib.metadata.version("investir")
    assert result.stdout.strip() == f"investir {version}"
    assert not result.stderr
    assert result.exit_code == EX_OK


def test_verbose_option(execute):
    result = execute(
        ["orders", DATA_FILE],
        global_opts=[
            "--verbose",
            "--offline",
            "--cache-file",
            os.devnull,
        ],
    )
    assert "INFO" in result.stderr
    assert "DEBUG" in result.stderr
    assert result.stdout
    assert result.exit_code == EX_OK


def test_default_verbosity(execute):
    result = execute(
        ["orders", DATA_FILE],
        global_opts=[
            "--offline",
            "--cache-file",
            os.devnull,
        ],
    )
    assert "INFO" in result.stderr
    assert "DEBUG" not in result.stderr
    assert result.stdout
    assert result.exit_code == EX_OK


@pytest.mark.network
@pytest.mark.skipif(
    sys.version_info < (3, 13),
    reason="Skipping test to avoid hitting API limits for Yahoo Finance",
)
def test_capital_gains_with_splits_downloaded_from_internet(execute):
    data_file = str(TEST_DIR / "orders_with_share_splits.csv")
    result = execute(
        ["capital-gains", data_file],
        global_opts=[
            "--quiet",
            # "--offline",  # Get splits from the Internet
            "--cache-file",
            os.devnull,
        ],
    )

    assert not result.stderr
    assert result.exit_code == EX_OK


@pytest.mark.network
@pytest.mark.skipif(
    sys.version_info < (3, 13),
    reason="Skipping test to avoid hitting API limits for Yahoo Finance",
)
def test_holdings_with_unrealised_gain_loss_calculated(execute):
    result = execute(
        ["holdings", "--show-gain-loss", "--ticker", "MSFT", DATA_FILE],
        global_opts=[
            "--quiet",
            "--cache-file",
            os.devnull,
        ],
    )
    assert "Not available" not in result.stdout
    assert not result.stderr
    assert result.exit_code == EX_OK


def test_parser_not_found_error(execute, tmp_path):
    csv_file = tmp_path / "transactions.csv"
    with csv_file.open("w", encoding="utf-8") as file:
        writer = csv.DictWriter(file, fieldnames=["Uknown"])
        writer.writeheader()

    result = execute(["orders", str(csv_file)])
    assert "Unable to find a parser for" in result.stderr
    assert not result.stdout
    assert result.exit_code != EX_OK


def test_total_amount_error(execute):
    csv_file = TEST_DIR / "total_amount_error.csv"
    result = execute(["orders", str(csv_file)])
    assert not result.stdout
    assert "Calculated amount" in result.stderr
    assert result.exit_code != EX_OK


def test_tax_calculator_error(execute):
    csv_file = TEST_DIR / "tax_calculator_error.csv"
    result = execute(["capital-gains", str(csv_file)])
    assert not result.stdout
    assert "Records appear to be incomplete" in result.stderr
    assert result.exit_code != EX_OK



import logging
from pathlib import Path
from typing import NamedTuple, Protocol

from investir.transaction import Dividend, Interest, Order, Transfer

logger = logging.getLogger(__name__)


class ParsingResult(NamedTuple):
    orders: list[Order]
    dividends: list[Dividend]
    transfers: list[Transfer]
    interest: list[Interest]


class Parser(Protocol):
    def __init__(self, csv_file: Path) -> None:
        pass

    @staticmethod
    def name() -> str:
        pass

    def can_parse(self) -> bool:
        pass

    def parse(self) -> ParsingResult:
        pass



import logging
from dataclasses import dataclass
from pathlib import Path

from platformdirs import user_cache_dir


@dataclass
class Config:
    strict: bool = True
    offline: bool = False
    cache_file: Path = Path(user_cache_dir()) / "investir" / "securities.yaml"
    include_fx_fees: bool = True
    log_level: int = logging.INFO
    use_colour: bool = True

    @property
    def logging_enabled(self) -> bool:
        return self.log_level != logging.CRITICAL

    def reset(self) -> None:
        Config.__init__(self)


config = Config()



import logging
from decimal import Decimal
from pathlib import Path
from typing import Final

import yaml
from iso4217 import Currency

from investir.findata.dataprovider import (
    DataProviderError,
    ExchangeRateProvider,
    SecurityInfoProvider,
)
from investir.findata.types import Price, SecurityInfo
from investir.trhistory import TrHistory
from investir.typing import ISIN

logger = logging.getLogger(__name__)


class FinancialData:
    VERSION: Final = 1

    def __init__(
        self,
        security_info_provider: SecurityInfoProvider | None,
        exchange_rate_provider: ExchangeRateProvider | None,
        tr_hist: TrHistory,
        cache_file: Path,
    ) -> None:
        self._security_info_provider = security_info_provider
        self._exchange_rate_provider = exchange_rate_provider
        self._tr_hist = tr_hist
        self._cache_file = cache_file
        self._security_info: dict[ISIN, SecurityInfo] = {}
        self._security_price: dict[ISIN, Price] = {}
        self._exchange_rates: dict[tuple[Currency, Currency], Decimal] = {}

        self._initialise()

    def get_security_info(self, isin: ISIN) -> SecurityInfo:
        return self._security_info[isin]

    def get_security_price(self, isin) -> Price | None:
        if price := self._security_price.get(isin):
            return price

        if self._security_info_provider is not None:
            try:
                price = self._security_info_provider.fetch_price(isin)
                logger.debug(
                    "Using %s %s share price for %s",
                    round(price.amount, 2),
                    price.currency.code,
                    isin,
                )
                self._security_price[isin] = price
            except DataProviderError as ex:
                logger.warning(str(ex))

        return price

    def get_foreign_exchange_rate(
        self, currency_from: Currency, currency_to: Currency
    ) -> Decimal | None:
        if fx_rate := self._exchange_rates.get((currency_from, currency_to)):
            return fx_rate

        if self._exchange_rate_provider is not None:
            try:
                fx_rate = self._exchange_rate_provider.fetch_exchange_rate(
                    currency_from, currency_to
                )
                inverse_fx_rate = Decimal("1.0") / fx_rate
                logger.debug(
                    "Using %s exchange rate for %s (%s) to %s (%s) (inverse rate: %s)",
                    round(fx_rate, 5),
                    currency_from.currency_name,
                    currency_from.code,
                    currency_to.currency_name,
                    currency_to.code,
                    round(inverse_fx_rate, 5),
                )
                self._exchange_rates[(currency_from), (currency_to)] = fx_rate
                self._exchange_rates[(currency_to), (currency_from)] = inverse_fx_rate
            except DataProviderError as ex:
                logger.warning(str(ex))

        return fx_rate

    def convert_currency(
        self,
        amount: Decimal,
        currency_from: Currency,
        currency_to: Currency = Currency.GBP,  # type: ignore[attr-defined]
    ) -> Decimal | None:
        if currency_from == currency_to:
            return amount

        if fx_rate := self.get_foreign_exchange_rate(currency_from, currency_to):
            return amount * fx_rate

        return None

    def _initialise(self) -> None:
        self._load_cache()

        orders = self._tr_hist.orders
        update_cache = False

        for isin, name in self._tr_hist.securities:
            if (security_info := self._security_info.get(isin)) is not None:
                last_order = next(o for o in reversed(orders) if o.isin == isin)

                if security_info.last_updated > last_order.timestamp:
                    logger.debug(
                        "Securities cache for %s (%s) is up-to-date", name, isin
                    )
                    continue

            self._security_info[isin] = SecurityInfo(name, [])
            update_cache = True

            if self._security_info_provider is not None:
                logger.info("Fetching information for %s - %s", isin, name)
                try:
                    self._security_info[isin] = self._security_info_provider.fech_info(
                        isin
                    )
                except DataProviderError as ex:
                    logger.warning(str(ex))

        if update_cache:
            self._update_cache()

    def _load_cache(self) -> None:
        if self._cache_file.exists():
            logger.info("Loading securities cache from %s", self._cache_file)

            with self._cache_file.open("r") as file:
                if data := yaml.load(file, Loader=yaml.FullLoader):
                    self._security_info = data["securities"]

    def _update_cache(self) -> None:
        if self._cache_file.exists():
            logger.info("Updating securities cache on %s", self._cache_file)
        else:
            logger.info("Creating securities cache on %s", self._cache_file)

        self._cache_file.parent.mkdir(parents=True, exist_ok=True)
        securities_info = dict(sorted(self._security_info.items()))
        data = {"version": self.__class__.VERSION, "securities": securities_info}

        with self._cache_file.open("w") as file:
            yaml.dump(data, file, sort_keys=False)



import csv
import logging
from collections.abc import Mapping
from decimal import ROUND_DOWN, Decimal
from pathlib import Path
from typing import Final

from dateutil.parser import parse as parse_timestamp

from investir.config import config
from investir.const import MIN_TIMESTAMP
from investir.exceptions import (
    CalculatedAmountError,
    CurrencyError,
    FeesError,
    OrderDateError,
    TransactionTypeError,
)
from investir.parser.factory import ParserFactory
from investir.parser.types import ParsingResult
from investir.transaction import (
    Acquisition,
    Disposal,
    Dividend,
    Interest,
    Order,
    Transfer,
)
from investir.typing import ISIN, Ticker
from investir.utils import dict2str, raise_or_warn, read_decimal

logger = logging.getLogger(__name__)


@ParserFactory.register("Freetrade")
class FreetradeParser:
    FIELDS: Final = (
        "Title",
        "Type",
        "Timestamp",
        "Account Currency",
        "Total Amount",
        "Buy / Sell",
        "Ticker",
        "ISIN",
        "Price per Share in Account Currency",
        "Stamp Duty",
        "Quantity",
        "Venue",
        "Order ID",
        "Order Type",
        "Instrument Currency",
        "Total Shares Amount",
        "Price per Share",
        "FX Rate",
        "Base FX Rate",
        "FX Fee (BPS)",
        "FX Fee Amount",
        "Dividend Ex Date",
        "Dividend Pay Date",
        "Dividend Eligible Quantity",
        "Dividend Amount Per Share",
        "Dividend Gross Distribution Amount",
        "Dividend Net Distribution Amount",
        "Dividend Withheld Tax Percentage",
        "Dividend Withheld Tax Amount",
    )

    def __init__(self, csv_file: Path) -> None:
        self._csv_file = csv_file
        self._orders: list[Order] = []
        self._dividends: list[Dividend] = []
        self._transfers: list[Transfer] = []
        self._interest: list[Interest] = []

    def can_parse(self) -> bool:
        with self._csv_file.open(encoding="utf-8") as file:
            reader = csv.DictReader(file)
            return tuple(reader.fieldnames or []) == self.FIELDS

    def parse(self) -> ParsingResult:
        parse_fn = {
            "ORDER": self._parse_order,
            "DIVIDEND": self._parse_dividend,
            "TOP_UP": self._parse_transfer,
            "WITHDRAWAL": self._parse_transfer,
            "INTEREST_FROM_CASH": self._parse_interest,
            "MONTHLY_STATEMENT": None,
            "TAX_CERTIFICATE": None,
        }

        with self._csv_file.open(encoding="utf-8") as file:
            reader = csv.DictReader(file)

            # Freetrade transactions are ordered from most recent to
            # oldest but we want the order ID to increase from the
            # oldest to the most recent.
            rows = reversed(list(reader))

            for row in rows:
                tr_type = row["Type"]

                if tr_type not in parse_fn:
                    raise_or_warn(TransactionTypeError(self._csv_file, row, tr_type))
                    continue

                if fn := parse_fn.get(tr_type):
                    if row["Account Currency"] != "GBP":
                        raise CurrencyError(
                            self._csv_file,
                            row,
                            row["Account Currency"],
                        )
                    fn(row)

        return ParsingResult(
            self._orders, self._dividends, self._transfers, self._interest
        )

    def _parse_order(self, row: Mapping[str, str]) -> None:
        title = row["Title"]
        timestamp = parse_timestamp(row["Timestamp"])
        total_amount = Decimal(row["Total Amount"])
        action = row["Buy / Sell"]
        ticker = row["Ticker"]
        isin = row["ISIN"]
        price = Decimal(row["Price per Share in Account Currency"])
        quantity = Decimal(row["Quantity"])
        order_id = row["Order ID"]
        stamp_duty = read_decimal(row["Stamp Duty"])
        fx_fee_amount = read_decimal(row["FX Fee Amount"])

        if action not in ("BUY", "SELL"):
            raise TransactionTypeError(self._csv_file, row, action)

        if timestamp < MIN_TIMESTAMP:
            raise OrderDateError(self._csv_file, row)

        if stamp_duty and fx_fee_amount:
            raise FeesError(self._csv_file, row)

        order_class: type[Order] = Acquisition
        fees = stamp_duty + fx_fee_amount

        if action == "SELL":
            order_class = Disposal
            fees *= -1

        calculated_amount = round(price * quantity + fees, 2)
        if calculated_amount != total_amount:
            raise_or_warn(
                CalculatedAmountError(
                    self._csv_file, row, total_amount, calculated_amount
                )
            )

        if config.include_fx_fees:
            allowable_fees = abs(fees)
        else:
            allowable_fees = stamp_duty

        self._orders.append(
            order_class(
                timestamp,
                isin=ISIN(isin),
                ticker=Ticker(ticker),
                name=title,
                amount=total_amount - fees,
                quantity=quantity,
                fees=allowable_fees,
                tr_id=order_id,
            )
        )

        logger.debug("Parsed row %s as %s\n", dict2str(row), self._orders[-1])

    def _parse_dividend(self, row: Mapping[str, str]):
        title = row["Title"]
        timestamp = parse_timestamp(row["Timestamp"])
        total_amount = Decimal(row["Total Amount"])
        ticker = row["Ticker"]
        isin = row["ISIN"]
        base_fx_rate = read_decimal(row["Base FX Rate"], Decimal("1.0"))
        eligible_quantity = Decimal(row["Dividend Eligible Quantity"])
        amount_per_share = Decimal(row["Dividend Amount Per Share"])
        withheld_tax_percentage = Decimal(row["Dividend Withheld Tax Percentage"])
        withheld_tax_amount = Decimal(row["Dividend Withheld Tax Amount"])

        calculated_ta = (
            amount_per_share
            * eligible_quantity
            * ((Decimal("100") - withheld_tax_percentage) / 100)
            * base_fx_rate
        )

        calculated_ta = calculated_ta.quantize(Decimal("1.00"), rounding=ROUND_DOWN)

        # Freetrade does not seem to use a consistent method for rounding dividends.
        # Thus, allow the calculated amount to differ by one pence.
        # https://community.freetrade.io/t/dividend-amount-off-by-one-penny/71806/7
        if abs(total_amount - calculated_ta) > Decimal("0.01"):
            raise_or_warn(
                CalculatedAmountError(self._csv_file, row, total_amount, calculated_ta)
            )

        self._dividends.append(
            Dividend(
                timestamp,
                isin=ISIN(isin),
                ticker=Ticker(ticker),
                name=title,
                amount=total_amount,
                withheld=withheld_tax_amount * base_fx_rate,
            )
        )

        logger.debug("Parsed row %s as %s\n", dict2str(row), self._dividends[-1])

    def _parse_transfer(self, row: Mapping[str, str]):
        timestamp = parse_timestamp(row["Timestamp"])
        tr_type = row["Type"]
        total_amount = Decimal(row["Total Amount"])

        if tr_type == "WITHDRAWAL":
            total_amount *= -1

        self._transfers.append(Transfer(timestamp, total_amount))

        logger.debug("Parsed row %s as %s\n", dict2str(row), self._transfers[-1])

    def _parse_interest(self, row: Mapping[str, str]):
        timestamp = parse_timestamp(row["Timestamp"])
        total_amount = Decimal(row["Total Amount"])

        self._interest.append(Interest(timestamp, total_amount))

        logger.debug("Parsed row %s as %s\n", dict2str(row), self._interest[-1])



import logging
from collections.abc import Callable
from pathlib import Path
from typing import ClassVar

from investir.parser.types import Parser

logger = logging.getLogger(__name__)


class ParserFactory:
    _parsers: ClassVar[dict[str, type[Parser]]] = {}

    @classmethod
    def register(cls, parser_name: str) -> Callable:
        def _wrapper(parser_class: type[Parser]) -> type[Parser]:
            cls._parsers[parser_name] = parser_class
            return parser_class

        return _wrapper

    @classmethod
    def create_parser(cls, csv_file: Path) -> Parser | None:
        for parser_name, parser_class in cls._parsers.items():
            parser = parser_class(csv_file)
            if parser.can_parse():
                logger.info("Found parser: %s", parser_name)
                return parser

        return None



from investir.findata.dataprovider import (
    DataProviderError,
    YahooFinanceExchangeRateProvider,
    YahooFinanceSecurityInfoProvider,
)
from investir.findata.financialdata import FinancialData
from investir.findata.types import Price, SecurityInfo, Split

__all__ = [
    "DataProviderError",
    "FinancialData",
    "Price",
    "SecurityInfo",
    "Split",
    "YahooFinanceExchangeRateProvider",
    "YahooFinanceSecurityInfoProvider",
]



from dataclasses import replace
from datetime import datetime
from decimal import Decimal

import pytest

from investir.exceptions import AmbiguousTickerError
from investir.transaction import Acquisition, Disposal, Dividend, Interest, Transfer
from investir.trhistory import TrHistory
from investir.typing import ISIN, Ticker

ORDER1 = Acquisition(
    datetime(2023, 4, 6, 18, 4, 50),
    isin=ISIN("AMZN-ISIN"),
    ticker=Ticker("AMZN"),
    name="Amazon",
    amount=Decimal("10.0"),
    quantity=Decimal("1.0"),
    fees=Decimal("0.5"),
    tr_id="ORDER1",
)

ORDER2 = Disposal(
    datetime(2024, 2, 5, 14, 7, 20),
    isin=ISIN("GOOG-ISIN"),
    ticker=Ticker("GOOG"),
    name="Alphabet",
    amount=Decimal("15.0"),
    quantity=Decimal("2.0"),
    fees=Decimal("1.0"),
    tr_id="ORDER2",
)

ORDER3 = Disposal(
    datetime(2025, 1, 2),
    isin=ISIN("AAPL-ISIN"),
    ticker=Ticker("AAPL"),
    name="Apple",
    amount=Decimal("1.0"),
    quantity=Decimal("1.0"),
    tr_id="ORDER3",
)

ORDER4 = Disposal(
    datetime(2025, 1, 3),
    isin=ISIN("AAPL-ISIN"),
    ticker=Ticker("AAPL"),
    name="Apple",
    amount=Decimal("1.0"),
    quantity=Decimal("1.0"),
    tr_id="ORDER4",
)

DIVIDEND1 = Dividend(
    datetime(2023, 2, 5, 14, 7, 20),
    isin=ISIN("AMZN-ISIN"),
    ticker=Ticker("AMZN"),
    name="Amazon",
    amount=Decimal("5.0"),
    withheld=Decimal("2.0"),
)

DIVIDEND2 = Dividend(
    datetime(2024, 2, 5, 14, 7, 20),
    isin=ISIN("GOOG-ISIN"),
    ticker=Ticker("GOOG"),
    name="Alphabet",
    amount=Decimal("5.0"),
    withheld=Decimal("2.0"),
)

TRANSFER1 = Transfer(datetime(2023, 2, 5, 14, 7, 20), Decimal("3000.0"))

TRANSFER2 = Transfer(datetime(2024, 2, 5, 14, 7, 20), Decimal("-1000.0"))

INTEREST1 = Interest(datetime(2023, 2, 5, 14, 7, 20), Decimal("1000.0"))

INTEREST2 = Interest(datetime(2024, 2, 5, 14, 7, 20), Decimal("500.0"))


def test_trhistory_duplicates_on_different_files_are_removed():
    # Create an order almost identical to ORDER1 other than the
    # `number` field which is automatically populated and it will be
    # different. For comparison purposes, the orders should be
    # equivalent.
    order1b = replace(ORDER1)

    tr_hist = TrHistory(
        orders=[ORDER1, order1b],
        dividends=[DIVIDEND1, DIVIDEND1],
        transfers=[TRANSFER1, TRANSFER1],
        interest=[INTEREST1, INTEREST1],
    )

    assert len(tr_hist.orders) == 1
    assert len(tr_hist.dividends) == 1
    assert len(tr_hist.transfers) == 1
    assert len(tr_hist.interest) == 1


def test_trhistory_transactions_are_sorted_by_timestamp():
    tr_hist = TrHistory(orders=[ORDER2, ORDER1])

    orders = tr_hist.orders
    assert len(orders) == 2
    assert orders[0].isin == ORDER1.isin
    assert orders[1].isin == ORDER2.isin


def test_trhistory_dividends_are_sorted_by_timestamp():
    tr_hist = TrHistory(dividends=[DIVIDEND2, DIVIDEND1])

    dividends = tr_hist.dividends
    assert len(dividends) == 2
    assert dividends[0].isin == DIVIDEND1.isin
    assert dividends[1].isin == DIVIDEND2.isin


def test_trhistory_transfers_are_sorted_by_timestamp():
    tr_hist = TrHistory(transfers=[TRANSFER2, TRANSFER1])

    transfers = tr_hist.transfers
    assert len(transfers) == 2
    assert transfers[0].amount == Decimal("3000")
    assert transfers[1].amount == Decimal("-1000")


def test_trhistory_interest_is_sorted_by_timestamp():
    tr_hist = TrHistory(interest=[INTEREST2, INTEREST1])

    interest = tr_hist.interest
    assert len(interest) == 2
    assert interest[0].amount == INTEREST1.amount
    assert interest[1].amount == INTEREST2.amount


def test_trhistory_securities():
    tr_hist = TrHistory(orders=[ORDER1, ORDER2, ORDER3, ORDER4])
    assert tuple(tr_hist.securities) == (
        ("GOOG-ISIN", "Alphabet"),
        ("AMZN-ISIN", "Amazon"),
        ("AAPL-ISIN", "Apple"),
    )


def test_get_security_name():
    tr_hist = TrHistory(orders=[ORDER1, ORDER2, ORDER3, ORDER4])
    assert tr_hist.get_security_name(ISIN("AMZN-ISIN")) == "Amazon"
    assert tr_hist.get_security_name(ISIN("GOOG-ISIN")) == "Alphabet"
    assert tr_hist.get_security_name(ISIN("AAPL-ISIN")) == "Apple"
    assert tr_hist.get_security_name(ISIN("NOTF")) is None


def test_get_ticker_isin():
    tr_hist = TrHistory(orders=[ORDER1, ORDER2, ORDER3, ORDER4])
    assert tr_hist.get_ticker_isin(Ticker("AMZN")) == ISIN("AMZN-ISIN")
    assert tr_hist.get_ticker_isin(Ticker("GOOG")) == ISIN("GOOG-ISIN")
    assert tr_hist.get_ticker_isin(Ticker("AAPL")) == ISIN("AAPL-ISIN")
    assert tr_hist.get_ticker_isin(Ticker("NOTF")) is None


def test_get_ticker_isin_when_ticker_ambigous():
    order1 = Acquisition(
        datetime(2023, 1, 1),
        isin=ISIN("NL0010273215"),
        ticker=Ticker("ASML"),
        amount=Decimal("10.0"),
        quantity=Decimal("1.0"),
    )

    order2 = Acquisition(
        datetime(2023, 1, 2),
        isin=ISIN("USN070592100"),
        ticker=Ticker("ASML"),
        amount=Decimal("10.0"),
        quantity=Decimal("1.0"),
    )

    tr_hist = TrHistory(orders=[order1, order2])
    with pytest.raises(AmbiguousTickerError):
        tr_hist.get_ticker_isin(Ticker("ASML"))


def test_trhistory_get_orders_table():
    tr_hist = TrHistory(orders=[ORDER1, ORDER2])
    table = tr_hist.get_orders_table()
    assert ORDER1.name in str(table)
    assert ORDER2.name in str(table)


def test_trhistory_get_dividends_table():
    tr_hist = TrHistory(dividends=[DIVIDEND1, DIVIDEND2])
    table = tr_hist.get_dividends_table()
    assert DIVIDEND1.name in str(table)
    assert DIVIDEND2.name in str(table)


def test_trhistory_get_transfers_table():
    tr_hist = TrHistory(transfers=[TRANSFER1, TRANSFER2])
    table = tr_hist.get_transfers_table()
    assert str(TRANSFER1.amount) in str(table)
    assert str(abs(TRANSFER2.amount)) in str(table)


def test_trhistory_get_interest_table():
    tr_hist = TrHistory(interest=[INTEREST1, INTEREST2])
    table = tr_hist.get_interest_table()
    assert str(INTEREST1.amount) in str(table)
    assert str(INTEREST2.amount) in str(table)



from collections.abc import Callable, Mapping, Sequence, ValuesView
from typing import NamedTuple, TypeVar

from investir.exceptions import AmbiguousTickerError
from investir.prettytable import PrettyTable
from investir.transaction import (
    Acquisition,
    Dividend,
    Interest,
    Order,
    Transaction,
    Transfer,
)
from investir.typing import ISIN, Ticker
from investir.utils import multifilter

T = TypeVar("T", bound=Transaction)


def unique_and_sorted(transactions: Sequence[T] | None) -> Sequence[T]:
    """Remove duplicated transactions and sort them by timestamp."""
    return sorted(set(transactions or []), key=lambda tr: tr.timestamp)


class Security(NamedTuple):
    isin: ISIN
    name: str = ""


class TrHistory:
    def __init__(
        self,
        *,
        orders: Sequence[Order] | None = None,
        dividends: Sequence[Dividend] | None = None,
        transfers: Sequence[Transfer] | None = None,
        interest: Sequence[Interest] | None = None,
    ) -> None:
        self._orders = unique_and_sorted(orders)
        self._dividends = unique_and_sorted(dividends)
        self._transfers = unique_and_sorted(transfers)
        self._interest = unique_and_sorted(interest)
        self._securities: Mapping[ISIN, Security] = {}

    @property
    def orders(self) -> Sequence[Order]:
        return self._orders

    @property
    def dividends(self) -> Sequence[Dividend]:
        return self._dividends

    @property
    def transfers(self) -> Sequence[Transfer]:
        return self._transfers

    @property
    def interest(self) -> Sequence[Interest]:
        return self._interest

    @property
    def securities(self) -> ValuesView[Security]:
        return self._securities_map().values()

    def get_security_name(self, isin: ISIN) -> str | None:
        security = self._securities_map().get(isin)
        return security.name if security else None

    def get_ticker_isin(self, ticker: Ticker) -> ISIN | None:
        isins = set(o.isin for o in self._orders if o.ticker == ticker)

        match len(isins):
            case 0:
                return None
            case 1:
                return next(iter(isins))
            case _:
                raise AmbiguousTickerError(ticker)

    def get_orders_table(
        self, filters: Sequence[Callable] | None = None
    ) -> PrettyTable:
        table = PrettyTable(
            field_names=(
                "Date",
                "Security Name",
                "ISIN",
                "Ticker",
                "Total Cost (£)",
                "Net Proceeds (£)",
                "Quantity",
                "Price (£)",
                "Fees (£)",
            ),
            show_total_fields=(
                "Total Cost (£)",
                "Net Proceeds (£)",
                "Fees (£)",
            ),
        )

        transactions = list(multifilter(filters, self._orders))
        last_idx = len(transactions) - 1

        for idx, tr in enumerate(transactions):
            net_proceeds = None
            total_cost = None
            if isinstance(tr, Acquisition):
                total_cost = tr.total_cost
            else:
                net_proceeds = tr.net_proceeds

            divider = (
                idx == last_idx or tr.tax_year() != transactions[idx + 1].tax_year()
            )

            table.add_row(
                [
                    tr.date,
                    tr.name,
                    tr.isin,
                    tr.ticker,
                    total_cost,
                    net_proceeds,
                    tr.quantity,
                    tr.price,
                    tr.fees,
                ],
                divider=divider,
            )

        return table

    def get_dividends_table(
        self, filters: Sequence[Callable] | None = None
    ) -> PrettyTable:
        table = PrettyTable(
            field_names=(
                "Date",
                "Security Name",
                "ISIN",
                "Ticker",
                "Net Amount (£)",
                "Widthheld Amount (£)",
            ),
            show_total_fields=(
                "Net Amount (£)",
                "Widthheld Amount (£)",
            ),
        )

        transactions = list(multifilter(filters, self._dividends))
        last_idx = len(transactions) - 1

        for idx, tr in enumerate(transactions):
            divider = (
                idx == last_idx or tr.tax_year() != transactions[idx + 1].tax_year()
            )

            table.add_row(
                [tr.date, tr.name, tr.isin, tr.ticker, tr.amount, tr.withheld or "n/a"],
                divider=divider,
            )

        return table

    def get_transfers_table(
        self, filters: Sequence[Callable] | None = None
    ) -> PrettyTable:
        table = PrettyTable(
            field_names=("Date", "Deposit (£)", "Withdrawal (£)"),
            show_total_fields=("Deposit (£)", "Withdrawal (£)"),
        )

        transactions = list(multifilter(filters, self._transfers))
        last_idx = len(transactions) - 1

        for idx, tr in enumerate(transactions):
            if tr.amount > 0:
                deposited = tr.amount
                widthdrew = ""
            else:
                deposited = ""
                widthdrew = abs(tr.amount)

            divider = (
                idx == last_idx or tr.tax_year() != transactions[idx + 1].tax_year()
            )

            table.add_row([tr.date, deposited, widthdrew], divider=divider)

        return table

    def get_interest_table(
        self, filters: Sequence[Callable] | None = None
    ) -> PrettyTable:
        table = PrettyTable(
            field_names=("Date", "Amount (£)"), show_total_fields=("Amount (£)",)
        )

        transactions = list(multifilter(filters, self._interest))
        last_idx = len(transactions) - 1

        for idx, tr in enumerate(transactions):
            divider = (
                idx == last_idx or tr.tax_year() != transactions[idx + 1].tax_year()
            )

            table.add_row([tr.date, tr.amount], divider=divider)

        return table

    def _securities_map(self) -> Mapping[ISIN, Security]:
        if not self._securities:
            self._securities = {
                o.isin: Security(o.isin, o.name)
                for o in sorted(self._orders, key=lambda o: o.name)
            }
        return self._securities
